import os
import sys
from pathlib import Path

# Add the project directory to the sys.path
project_dir = str(Path(__file__).resolve().parents[2])
if project_dir not in sys.path:
    sys.path.append(project_dir)
from dotenv import load_dotenv
load_dotenv()
import aiohttp
import pandas as pd
from pytz import timezone

from .trade_models.capital_flow import CapitalFlow
from .trade_models.cost_distribution import CostDistribution
from .trade_models.etf_holdings import ETFHoldings
from .trade_models.institutional_holdings import InstitutionHolding, InstitutionStat
from .trade_models.financials import BalanceSheet, FinancialStatement, CashFlow
from .trade_models.news import NewsItem
from .trade_models.forecast_evaluator import ForecastEvaluator
from .trade_models.short_interest import ShortInterest
from .trade_models.volume_analysis import WebullVolAnalysis
from .trade_models.ticker_query import WebullStockData, MultiQuote
from .trade_models.analyst_ratings import Analysis


from datetime import datetime, timedelta

class WebullTrading:
    def __init__(self):


        self.today = datetime.now().strftime('%Y-%m-%d')
        self.yesterday = (datetime.now() - timedelta(days=1)).strftime('%Y-%m-%d')
        self.tomorrow = (datetime.now() + timedelta(days=1)).strftime('%Y-%m-%d')
        self.thirty_days_ago = (datetime.now() - timedelta(days=30)).strftime('%Y-%m-%d')
        self.thirty_days_from_now = (datetime.now() + timedelta(days=30)).strftime('%Y-%m-%d')
        self.fifteen_days_ago = (datetime.now() - timedelta(days=15)).strftime('%Y-%m-%d')
        self.fifteen_days_from_now = (datetime.now() + timedelta(days=15)).strftime('%Y-%m-%d')
        self.eight_days_from_now = (datetime.now() + timedelta(days=8)).strftime('%Y-%m-%d')
        self.eight_days_ago = (datetime.now() - timedelta(days=8)).strftime('%Y-%m-%d')
        self.timeframes = ['m1','m5', 'm10', 'm15', 'm20', 'm30', 'm60', 'm120', 'm240', 'd1']

        self.headers = {
        "Access_token": os.environ.get('ACCESS_TOKEN'),
        "App": "global",
        "App-Group": "broker",
        "Appid": "wb_web_app",
        "Content-Type": "application/json",
        "Device-Type": "Web",
        "Did": os.environ.get('DID'),
        "Hl": "en",
        "Locale": "eng",
        "Os": "web",
        "Osv": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/118.0.0.0 Safari/537.36",
        "Ph": "Windows Chrome",
        "Platform": "web",
        "Referer": "https://app.webull.com/",
        "Reqid": os.environ.get('REQ_ID'),
        "Sec-Ch-Ua": "\"Chromium\";v=\"118\", \"Google Chrome\";v=\"118\", \"Not=A?Brand\";v=\"99\"",
        "Sec-Ch-Ua-Mobile": "?0",
        "Sec-Ch-Ua-Platform": "\"Windows\"",
        "T_time": "1698276695206",
        "Tz": "America/Los_Angeles",
        "User-Agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/118.0.0.0 Safari/537.36",
        "Ver": "3.40.11",
        "X-S": "49ef20ad66d1e24a83ff8b2015bc13c6d133285c5665dbbe4aa6032572749931",
        "X-Sv": "xodp2vg9"
    }
    async def fetch_endpoint(self, endpoint):
        async with aiohttp.ClientSession(headers=self.headers) as session:
            async with session.get(endpoint) as resp:
                return await resp.json()

    async def get_ticker_id(self, symbol):
        """Converts ticker name to ticker ID to be passed to other API endpoints from Webull."""
        endpoint =f"https://quotes-gw.webullfintech.com/api/search/pc/tickers?keyword={symbol}&pageIndex=1&pageSize=1"

        
        data =  await self.fetch_endpoint(endpoint)
        datas = data['data'] if 'data' in data else None
        if datas is not None:
            tickerID = datas[0]['tickerId']
            return tickerID
    

    async def get_bars(self, symbol, timeframe:str='m1'):
        """
        Timeframes:
        
        >>> m1: 1 minute
        >>> m5: 5 minute
        >>> m10: 10 minute
        >>> m15: 15 minute
        >>> m20: 20 minute
        >>> m30: 30 minute
        >>> m60: 1 hour
        >>> m120: 2 hour
        >>> m240: 4 hour
        
        """
        tickerid = await self.get_ticker_id(symbol)
        endpoint = f"https://quotes-gw.webullfintech.com/api/quote/charts/query?tickerIds={tickerid}&type={timeframe}&count=800&extendTrading=0"
        datas =  await self.fetch_endpoint(endpoint)
        if datas is not None:
            data = datas[0]['data']
            # Create empty lists for each column
            timestamps = []
            column2 = []
            column3 = []
            column4 = []
            column5 = []
            column6 = []

            # Split each line and append values to respective lists
            for line in data:
                parts = line.split(',')
                timestamps.append(parts[0])
                column2.append(parts[1])
                column3.append(parts[2])
                column4.append(parts[3])
                column5.append(parts[4])
                column6.append(parts[5])

  

            df = pd.DataFrame({
                'Timestamp': timestamps,
                'Open': column2,
                'Low': column3,
                'High': column4,
                'Close': column5,
                'Vwap': column6
            })

            # Convert the 'Timestamp' column to integers before converting to datetime
            df['Timestamp'] = df['Timestamp'].astype(int)

            # Then convert to datetime
            df['Timestamp'] = pd.to_datetime(df['Timestamp'], unit='s')


            # Convert to Eastern Time
            eastern = timezone('US/Eastern')
            df['Timestamp'] = df['Timestamp'].dt.tz_localize('UTC').dt.tz_convert(eastern)

            # Remove the timezone information
            df['Timestamp'] = df['Timestamp'].dt.tz_localize(None)
            df['Timeframe'] = timeframe
            df['Ticker'] = symbol
  
            return df
        


    async def get_stock_quote(self, symbol:str):
        ticker_id = await self.get_ticker_id(symbol)

        endpoint = f"https://quotes-gw.webullfintech.com/api/stock/tickerRealTime/getQuote?tickerId={ticker_id}&includeSecu=1&includeQuote=1&more=1"
        datas = await self.fetch_endpoint(endpoint)

        data = WebullStockData(datas)
        return data


    async def get_analyst_ratings(self, symbol:str):
        ticker_id = await self.get_ticker_id(symbol)
        endpoint=f"https://quotes-gw.webullfintech.com/api/information/securities/analysis?tickerId={ticker_id}"
        datas = await self.fetch_endpoint(endpoint)
        data = Analysis(datas)
        return data
    

    async def get_short_interest(self, symbol:str):
        ticker_id = await self.get_ticker_id(symbol)
        endpoint = f"https://quotes-gw.webullfintech.com/api/information/brief/shortInterest?tickerId={ticker_id}"
        datas = await self.fetch_endpoint(endpoint)
        data = ShortInterest(datas)
        return data
    
    async def institutional_holding(self, symbol:str):
        ticker_id = await self.get_ticker_id(symbol)
        endpoint = f"https://quotes-gw.webullfintech.com/api/information/stock/getInstitutionalHolding?tickerId={ticker_id}"
        datas = await self.fetch_endpoint(endpoint)
        data = InstitutionStat(datas)

        return data
    

    async def volume_analysis(self, symbol:str):
        ticker_id = await self.get_ticker_id(symbol)
        endpoint = f"https://quotes-gw.webullfintech.com/api/stock/capitalflow/stat?count=10&tickerId={ticker_id}&type=0"
        datas = await self.fetch_endpoint(endpoint)
        data = WebullVolAnalysis(datas)
        return data
    

    async def cost_distribution(self, symbol:str, start_date:str=None, end_date:str=None):

        if start_date is None:
            start_date = self.yesterday
            

        if end_date is None:
            end_date = self.today

        ticker_id = await self.get_ticker_id(symbol)
        endpoint = f"https://quotes-gw.webullfintech.com/api/quotes/chip/query?tickerId={ticker_id}&startDate={start_date}&endDate={end_date}"
 
        datas = await self.fetch_endpoint(endpoint)
        data = CostDistribution(datas)
        return data
    

    async def stock_quote(self, symbol:str):
        ticker_id = await self.get_ticker_id(symbol)
        endpoint = f"https://quotes-gw.webullfintech.com/api/bgw/quote/realtime?ids={ticker_id}&includeSecu=1&delay=0&more=1"
        datas = await self.fetch_endpoint(endpoint)
        data = WebullStockData(datas)
        return data
    

    async def news(self, symbol:str, pageSize:str='100'):
        ticker_id = await self.get_ticker_id(symbol)
        endpoint = f"https://nacomm.webullfintech.com/api/information/news/tickerNews?tickerId={ticker_id}&currentNewsId=0&pageSize={pageSize}"
        datas = await self.fetch_endpoint(endpoint)
        data = NewsItem(datas)
        return data
    

    async def balance_sheet(self, symbol:str, limit:str='11'):
        ticker_id = await self.get_ticker_id(symbol)
        endpoint = f"https://quotes-gw.webullfintech.com/api/information/financial/balancesheet?tickerId={ticker_id}&type=101&fiscalPeriod=0&limit={limit}"
        datas = await self.fetch_endpoint(endpoint)
        data = BalanceSheet(datas)
        return data
    
    async def cash_flow(self, symbol:str, limit:str='12'):
        ticker_id = await self.get_ticker_id(symbol)
        endpoint = f"https://quotes-gw.webullfintech.com/api/information/financial/cashflow?tickerId={ticker_id}&type=102&fiscalPeriod=1,2,3,4&limit={limit}"
        datas = await self.fetch_endpoint(endpoint)
        data = CashFlow(datas)
        return data
    
    async def income_statement(self, symbol:str, limit:str='12'):
        ticker_id = await self.get_ticker_id(symbol)
        endpoint = f"https://quotes-gw.webullfintech.com/api/information/financial/incomestatement?tickerId={ticker_id}&type=102&fiscalPeriod=1,2,3,4&limit={limit}"
        datas = await self.fetch_endpoint(endpoint)
        data = FinancialStatement(datas)
        return data
    



    async def capital_flow(self, symbol:str):
        ticker_id = await self.get_ticker_id(symbol)
        endpoint = f"https://quotes-gw.webullfintech.com/api/stock/capitalflow/ticker?tickerId={ticker_id}&showHis=true"
        datas = await self.fetch_endpoint(endpoint)
        data = CapitalFlow(datas)
        return data
    

    async def etf_holdings(self, symbol:str, pageSize:str='200'):
        ticker_id = await self.get_ticker_id(symbol)
        endpoint = f"https://quotes-gw.webullfintech.com/api/information/company/queryEtfList?tickerId={ticker_id}&pageIndex=1&pageSize={pageSize}"
        datas = await self.fetch_endpoint(endpoint)
        data = ETFHoldings(datas)
        return data
    

    async def multi_quote(self, symbols:str):
        counter = 0
        while True:
            counter = counter + 1
           
            ticker_ids = [await self.get_ticker_id(i) for i in symbols]
            ticker_ids = str(ticker_ids)
            ticker_ids = ','.join([ticker_ids]).replace(']', '').replace('[', '').replace(' ', '')
            endpoint = f"https://quotes-gw.webullfintech.com/api/bgw/quote/realtime?ids={ticker_ids}&includeSecu=1&delay=0&more=1"

            async with aiohttp.ClientSession(headers=self.headers) as session:
                async with session.get(endpoint) as resp:
                    datas = await resp.json()
                    all_data = MultiQuote(datas)

                    for sym,price, vol, vr in zip(all_data.symbol, all_data.close, all_data.volume, all_data.vibrateRatio):
                        yield(f'SYM(1): | {sym} | PRICE(3): | {price} | VOL:(5): | {vol} | VIBRATION:(7): | {vr}')

                        if counter == 250:
                            print(f"Stream ending...")
                        
                            break


import pandas as pd
import pytz
from datetime import datetime
import asyncpg
from asyncpg.exceptions import InvalidCatalogNameError
import os

from dataclasses import dataclass, asdict
from typing import List


@dataclass
class AggregatesData:
    def __init__(self, response_data):
        ticker: str
        open: float
        high: float
        low: float
        close: float
        timestamp: str

    
        utc_tz = pytz.timezone('UTC')
        central_tz = pytz.timezone('America/Chicago')
        self.pool = None
        self.conn = os.environ.get('AGGREGATES_STRING')
        self.user = os.environ.get('DB_USER')
        self.host = "localhost"
        self.port = 5432

        self.adjusted = response_data.get("adjusted")
        self.next_url = response_data.get("next_url")
        self.ticker = response_data.get("ticker")
        self.query_count = response_data.get("queryCount")
        self.request_id = response_data.get("request_id")
        self.results = response_data.get("results")
        self.db_config = { 
            'user': os.environ.get('DB_USER'),
            'host': os.environ.get('DB_HOST'),
            'password': os.environ.get('DB_PASSWORD'),
            'port': os.environ.get('DB_PORT'),
            'name': os.environ.get('AGGS_DB')

        }
        if self.results is None:
            self.close = []
            self.high = []
            self.low = []
            self.n = []
            self.open = []
            self.timestamp = []
            self.volume = []
            self.volume_weighted_average = []
        else:
            self.close = [i['c'] for i in self.results]
            self.high = [i['h'] for i in self.results]
            self.low = [i['l'] for i in self.results]
      
            self.open = [i['o'] for i in self.results]
            self.timestamp = [datetime.utcfromtimestamp(i['t'] / 1000).replace(tzinfo=utc_tz).astimezone(central_tz).replace(tzinfo=None) for i in self.results]

            self.volume = [i['v'] for i in self.results] if not self.ticker.startswith("I:") else None


            self.data_dict = { 
                'Open': self.open,
                'High': self.high,
                'Low': self.low,
                'Close': self.close,
                
                'Timestamp': self.timestamp

            }


            self.df = pd.DataFrame(self.data_dict)

    def __str__(self):
        return f"AggregatesData({self.ticker})"

    def __repr__(self):
        return self.__str__()
    
    def to_db_format(self):
        return [(self.ticker, o, h, l, c, t) for o, h, l, c, t in zip(self.open, self.high, self.low, self.close, self.timestamp)]

    async def create_pool(self, db_config):
        """
        Create a connection pool using the given database configuration.

        Parameters:
        - db_config (dict): Database configuration parameters including user, password, database, host, port.

        Returns:
        - asyncpg.Pool: A connection pool.
        """
        self.pool = await asyncpg.create_pool(
            user=db_config['DB_USER'],
            password=db_config['DB_PASSWORD'],
            database=db_config['DB_NAME'],
            host=db_config['DB_HOST'],
            port=db_config['DB_PORT']
        )
        return self.pool

    # Function to create PostgreSQL table
  
    async def connect(self, conn_str):
        """
        Create a single connection to the database using the given connection string.

        Parameters:
        - conn_str (str): The database connection string.

        Returns:
        - asyncpg.Connection: A single database connection.
        """
        try:
            self.conn = await asyncpg.connect(conn_str)
        except InvalidCatalogNameError:
            print(f'Database does not exist! Attempting to create database...')
            
            # Extract database name from the connection string
            db_name = conn_str.split('/')[-1].split('?')[0]
            
            # Create a temporary connection string without the database name
            temp_conn_str = conn_str.replace(f"/{db_name}", "/postgres")
            
            # Connect to the default "postgres" database to create a new database
            temp_conn = await asyncpg.connect(temp_conn_str)
            
            try:
                await temp_conn.execute(f'CREATE DATABASE {db_name}')
                print(f"Database {db_name} created successfully.")
                
                # Now connect to the newly created database
                self.conn = await asyncpg.connect(conn_str)
                
            except Exception as e:
                print(f"Failed to create database. Error: {e}")
            finally:
                await temp_conn.close()
                
        return self.conn


    async def create_table(self):
    
        async with self.pool.acquire() as conn:
            async with conn.transaction():
                await conn.execute("""
                CREATE TABLE IF NOT EXISTS aggregates_data (
                    ticker VARCHAR(10),
                    open FLOAT,
                    high FLOAT,
                    low FLOAT,
                    close FLOAT,
                    timestamp TIMESTAMPTZ
                );
                """)
    def to_dict(self):
        return asdict(self)
    
    #function to insert into Postgres
    async def insert_data(self, data_dicts: List[dict]):
        async with self.pool.acquire() as conn:
            async with conn.transaction():
                for data_dict in data_dicts:
                    await conn.execute("""
                    INSERT INTO aggregates_data (ticker, open, high, low, close, timestamp)
                    VALUES ($1, $2, $3, $4, $5, $6);
                    """, data_dict['Ticker'], data_dict['Open'], data_dict['High'], data_dict['Low'], data_dict['Close'], data_dict['Timestamp'])


import disnake
from disnake.ext import commands
import pandas as pd
from tabulate import tabulate
from webull_options.webull_options import WebullOptions
from discord_.bot_menus.pagination import AlertMenus
import os

from dotenv import load_dotenv
load_dotenv()
theta_em = "<a:_:1047010360554704977>"
gamma_em = "<a:_:1190024505700134973>"
delta_em = "<a:_:1044647851466182736>"
vega_em = "<a:_:1044647903009976372>"
rho_em = "<a:_:1044647745702596648>"
ticker_em = "<:_:1190025407815221270>"
volume_em = "🔊"
oi_em = "👥"
oi_change_em = "🔀"
iv_em="💉"
expiry_em="🗓️"
strike_em="🎳"
call_put_em="↕️"
change_ratio_em="➗"
start_over_em="<:_:1190026332248232126>"
open_em="<a:_:1044647404693106728>"
high_em="<a:_:1047010108930019340>"
low_em="<a:_:1044658254137008148>"
close_em="<a:_:1044647558133334126>"
underlying_price_em="💲"
sdk = WebullOptions(access_token=os.environ.get('ACCESS_TOKEN'), osv=os.environ.get('OSV'), did=os.environ.get('DID'))
from disnake import TextInputStyle
class MyModal(disnake.ui.Modal):
    def __init__(self, label, placeholder, name, conn, query, strat=None):
        self.strat=strat
        self.label=label
        self.placeholder=placeholder
        self.name=name
        self.query=query
        self.conn=conn
        # The details of the modal, and its components
        components = [

            disnake.ui.TextInput(
                label='Operator',
                placeholder='Choose between >=, <=, >, <, or =',
                custom_id='operator',
                style=TextInputStyle.short,
                max_length=2

            ),
            disnake.ui.TextInput(
                label=self.label,
                placeholder=self.placeholder,
                custom_id=self.name,
                style=TextInputStyle.short,
                
            ),
        
        ]
        super().__init__(title="Create Tag", components=components)

    # The callback received when the user input is completed.
    async def callback(self, inter: disnake.ModalInteraction):
        user_input = inter.text_values

        value = user_input.get(self.name)
        operator = user_input.get('operator')

        select = StrategySelect()
        embeds = await select.get_results(self.conn, self.query, self.strat)

        await inter.edit_original_message(embeds=embeds[0], view=AlertMenus(embeds).add_item(StrategyView()))

class ExpiryButton(disnake.ui.Button):
    def __init__(self, query):
        self.query=query
     

        super().__init__( 
            style=disnake.ButtonStyle.blurple,
            label='Expiry',
            emoji=expiry_em,
            custom_id='expiry_button'
        )


    async def callback(self, inter:disnake.AppCmdInter):
        name = 'expire_date'
        await inter.response.send_modal(MyModal(label='Expiration Date', placeholder='e.g. 2024-03-15', name='expire_date', conn=await sdk.db_manager.get_connection(), query=self.query))


        

        

class StrategyView(disnake.ui.View):
    def __init__(self):
        super().__init__(timeout=None)


        self.add_item(StrategySelect())



class StrategySelect(disnake.ui.Select):
    def __init__(self):
        super().__init__( 
            custom_id='strategyselect',
            min_values=1,
            max_values=1,
            options=[ 
                disnake.SelectOption(label='Easily Managed', description=f'Low theta with time on the clock. Easily managed.'),
                disnake.SelectOption(label='Change Percent', description=f'Options that are down significantly with time left.'),
                disnake.SelectOption(label='Long Puts', description=f'Put options gain as stocks fall, with low time decay.'),
                disnake.SelectOption(label='OTM Puts', description=f'Put options slightly out of the money with moderate delta.')
            ]
        )


    def chunk_string(self, string, size):
        """Yield successive size-sized chunks from string."""
        for i in range(0, len(string), size):
            yield string[i:i + size]


    async def get_results(self, conn, query, strat):
        results = await conn.fetch(query)
        df = pd.DataFrame(results, columns = ['underlying_symbol', 'strike_price', 'call_put', 'expire_date', 'price'])
        print(df.columns)
        # Select only the specified columns
  
        df = df.rename(columns={'underlying_symbol': 'sym', 'strike_price':'strike', 'call_put': 'cp', 'expire_date': 'exp'})
        table = tabulate(df, headers=['sym', 'strike', 'cp', 'expiry', 'price'], tablefmt='fancy', showindex=False)
        # Break apart data into chunks of 4000 characters
        chunks = self.chunk_string(table, 4000)
        embeds=[]
        # Create and send embeds for each chunk
        # Strategy descriptions
        strategy_dict = {
            # Existing strategies
            'Easily Managed': 'Options with low theta for minimal time decay, ideal for stable strategies.  This query sifts through the options database to uncover potential trading opportunities with a short to medium-short time horizon. Its tailored for traders who prefer to engage in strategies that have time on the clock - with low theta decay - allowing for easy managability.',
            'Long Calls': 'Options with high delta, signaling strong potential price increase with the underlying stock.',
            'Long Puts': 'Put options that gain value as the underlying stock price falls, with low time decay.',
            # Add the new strategy
            'Moderate Out-of-the-Money Puts': 'Put options slightly out of the money with moderate delta, balancing risk and potential inverse movement to the stock.'
            # Add more strategies and their descriptions here
        }

        # [Earlier code]
        for chunk in chunks:
            embed = disnake.Embed(title=f"Strategy Menu | {strat}", description=f"```py\n{chunk}```")
            # Dynamically adding strategy information from the dictionary
            strategy_info = strategy_dict.get(strat, "No information available")
            embed.add_field(name="Strategy Info:", value=f"```py\n{strategy_info}```")
            embed.set_footer(text='Implemented by FUDSTOP')
            embeds.append(embed)
        return embeds


    async def callback(self, inter:disnake.AppCmdInter):
        conn = await sdk.db_manager.get_connection()
        await inter.response.defer()
        if self.values[0] == 'Easily Managed':
            query = f"""SELECT underlying_symbol, strike_price, expire_date, call_put, close 
                        FROM options 
                        WHERE expire_date >= '2024-01-01' AND expire_date <= '2025-12-31'
                        AND close >= 0 and theta <= -0.03;
                        """
            embeds = await self.get_results(conn,query, self.values[0])
            await inter.edit_original_message(embed=embeds[0], view=AlertMenus(embeds).add_item(StrategySelect()))


            return query

        if self.values[0] == 'Long Puts':
            query = """SELECT underlying_symbol, strike_price, expire_date, call_put, delta, theta 
                    FROM options 
                    WHERE call_put = 'put' AND delta <= -0.7 AND theta <= 0.02 
                    AND expire_date >= '2023-03-15';"""
            embeds = await self.get_results(conn,query, self.values[0])
            await inter.edit_original_message(embed=embeds[0], view=AlertMenus(embeds).add_item(StrategySelect()))


            return query
        

        if self.values[0] == 'High Delta / Low Theta':
            query = """SELECT underlying_symbol, strike_price, expire_date, call_put, delta, theta
           FROM options
           WHERE delta >= 0.7 AND theta <= 0.01 AND expire_date >= CURRENT_DATE + INTERVAL '5 months'
           ORDER BY delta DESC, theta ASC LIMIT 25;"""
            
            embeds = await self.get_results(conn,query, self.values[0])
            await inter.edit_original_message(embed=embeds[0], view=AlertMenus(embeds).add_item(StrategySelect()))


            return query
        
        if self.values[0] == 'OTM Puts':
            query = f"""SELECT underlying_symbol, strike_price, expire_date, call_put, delta
                        FROM options
                        WHERE call_put = 'put' AND delta <= -0.05
                        AND strike_price >= close + 5
                        LIMIT 25;
                        """
            
            embeds = await self.get_results(conn,query, self.values[0])
            await inter.edit_original_message(embed=embeds[0], view=AlertMenus(embeds).add_item(StrategySelect()))


            return query


class StrategyCOG(commands.Cog):
    def __init__(self, bot):
        self.bot=bot
        

    @commands.slash_command()
    async def strategy(self, inter):
        pass


    @strategy.sub_command()
    async def menu(self, inter:disnake.AppCmdInter):
        await inter.response.defer()

        await inter.edit_original_message(view=StrategyView())



def setup(bot:commands.Bot):
    bot.add_cog(StrategyCOG(bot))
    print('Strategy! Ready!')


import os
from dotenv import load_dotenv
load_dotenv()
import math
from scipy.stats import norm
from datetime import datetime

import numpy as np
import asyncio
from fudstop.apis.polygonio.polygon_options import PolygonOptions
from fudstop.list_sets.ticker_lists import most_active_tickers
from fudstop.apis.webull.webull_options import WebullOptions
from fudstop.apis.webull.webull_trading import WebullTrading
db_config = {
    "host": os.environ.get('DB_HOST', 'localhost'), # Default to this IP if 'DB_HOST' not found in environment variables
    "port": int(os.environ.get('DB_PORT')), # Default to 5432 if 'DB_PORT' not found
    "user": os.environ.get('DB_USER', 'postgres'), # Default to 'postgres' if 'DB_USER' not found
    "password": os.environ.get('DB_PASSWORD', 'fud'), # Use the password from environment variable or default
    "database": os.environ.get('DB_NAME', 'polygon') # Database name for the new jawless database
}
opts = WebullOptions(connection_string=os.environ.get('WEBULL_OPTIONS'))
poly = PolygonOptions()
wt = WebullTrading()
import pandas as pd
import aiohttp
import requests

######### YOU MUST EXPAND HERE

async def main():
    counter = 0
    async for quote in wt.multi_quote(['CHWY', 'M', 'ENPH', 'MRNA', 'JD']):
        counter += 1
        quote = quote.split('|')
        sym = quote[1]
        price = quote[3]
        vol = quote[5]
        vibration = quote[7]

        print(f"{sym} | ${price} | {vol} | {vibration}")
        # Perform a specific action every 300 iterations
        if counter % 300 == 0:
            # Perform some action here
            print("Action performed after 300 iterations.")

        # The loop will continue after the action

asyncio.run(main())



import pandas as pd

class OptionSnapshotData:
    def __init__(self, data):
        self.implied_volatility = [float(i['implied_volatility']) if 'implied_volatility' in i else None for i in data]
        self.open_interest = [float(i['open_interest']) if 'open_interest' in i else None for i in data]
        self.break_even_price = [float(i['break_even_price']) if 'break_even_price' in i else None for i in data]

        day = [i['day'] if i['day'] is not None else None for i in data]
        self.day_close = [float(i['close']) if 'close' in i else None for i in day]
        self.day_high = [float(i['high']) if 'high' in i else None for i in day]
        self.last_updated  = [i['last_updated'] if 'last_updated' in i else None for i in day]
        self.day_low  = [float(i['low']) if 'low' in i else None for i in day]
        self.day_open  = [float(i['open']) if 'open' in i else None for i in day]
        self.day_change_percent  = [float(i['change_percent']) if 'change_percent' in i else None for i in day]
        self.day_change  = [float(i['change']) if 'change' in i else None for i in day]
        self.previous_close = [float(i['previous_close']) if 'previous_close' in i else None for i in day]
        self.day_volume = [float(i['volume']) if 'volume' in i else None for i in day]
        self.day_vwap  = [float(i['vwap']) if 'vwap' in i else None for i in day]

        details = [i['details'] if i['details'] is not None else None for i in data]
        self.contract_type = [i['contract_type'] if 'contract_type' in i else None for i in details]
        self.exercise_style = [i['exercise_style'] if 'exercise_style' in i else None for i in details]
        self.expiration_date = [i['expiration_date'] if 'expiration_date' in i else None for i in details]
        self.shares_per_contract= [i['shares_per_contract'] if 'shares_per_contract' in i else None for i in details]
        self.strike_price = [float(i['strike_price']) if 'strike_price' in i else None for i in details]
        self.option_symbol = [i['ticker'] if 'ticker' in i else None for i in details]

        greeks = [i['greeks'] if i['greeks'] is not None else None for i in data]
        self.delta = [float(i['delta']) if 'delta' in i else None for i in greeks]
        self.gamma= [float(i['gamma']) if 'gamma' in i else None for i in greeks]
        self.theta= [float(i['theta']) if 'theta' in i else None for i in greeks]
        self.vega = [float(i['vega']) if 'vega' in i else None for i in greeks]

        lastquote = [i['last_quote'] if i['last_quote'] is not None else None for i in data]
        self.ask = [float(i['ask']) if 'ask' in i else None for i in lastquote]
        self.ask_size = [float(i['ask_size']) if 'ask_size' in i else None for i in lastquote]
        self.bid= [float(i['bid']) if 'bid' in i else None for i in lastquote]
        self.bid_size= [float(i['bid_size']) if 'bid_size' in i else None for i in lastquote]
        self.quote_last_updated= [i['quote_last_updated'] if 'quote_last_updated' in i else None for i in lastquote]
        self.midpoint = [float(i['midpoint']) if 'midpoint' in i else None for i in lastquote]


        lasttrade = [i['last_trade'] if i['last_trade'] is not None else None for i in data]
        self.conditions = [i['conditions'] if 'conditions' in i else None for i in lasttrade]
        self.exchange = [i['exchange'] if 'exchange' in i else None for i in lasttrade]
        self.price= [float(i['price']) if 'price' in i else None for i in lasttrade]
        self.sip_timestamp= [i['sip_timestamp'] if 'sip_timestamp' in i else None for i in lasttrade]
        self.size= [float(['size']) if 'size' in i else None for i in lasttrade]

        underlying = [i['underlying_asset'] if i['underlying_asset'] is not None else None for i in data]
        self.change_to_break_even = [i['change_to_break_even'] if 'change_to_break_even' in i else None for i in underlying]
        self.underlying_last_updated = [i['underlying_last_updated'] if 'underlying_last_updated' in i else None for i in underlying]
        self.underlying_price = [float(i['price']) if 'price' in i else None for i in underlying]
        self.underlying_ticker = [i['ticker'] if 'ticker' in i else None for i in underlying]


        self.data_dict = {
        "implied_volatility": self.implied_volatility,
        "open_interest": self.open_interest,
        "break_even_price": self.break_even_price,
        "close": self.day_close,
        "high": self.day_high,
        "last_updated": self.last_updated,
        "low": self.day_low,
        "open": self.day_open,
        "change_percent": self.day_change_percent,
        "change": self.day_change,
        "previous_close": self.previous_close,
        "vol": self.day_volume,
        "vwap": self.day_vwap,
        "call_put": self.contract_type,
        "exercise_style": self.exercise_style,
        "exp": self.expiration_date,
        "shares_per_contract": self.shares_per_contract,
        "strike": self.strike_price,
        "ticker": self.option_symbol,

        "delta": self.delta,
        "gamma": self.gamma,
        "theta": self.theta,
        "vega": self.vega,
        "ask": self.ask,
        "ask_size": self.ask_size,
        "bid": self.bid,
        "bid_size": self.bid_size,
        "quote_last_updated": self.quote_last_updated,
        "midpoint": self.midpoint,
        "conditions": self.conditions,
        "exchange": self.exchange,
        "cost": self.price,
        "sip_timestamp": self.sip_timestamp,
        "size": self.size,
        "change_to_break_even": self.change_to_break_even,
        "underlying_last_updated": self.underlying_last_updated,
        "price": self.underlying_price,
        "symbol": self.underlying_ticker
    }


        self.df = pd.DataFrame(self.data_dict)



import sys
from pathlib import Path

# Add the project directory to the sys.path
project_dir = str(Path(__file__).resolve().parents[2])
if project_dir not in sys.path:
    sys.path.append(project_dir)

import os

from dotenv import load_dotenv
load_dotenv()


import asyncio



from apis.polygonio.polygon_options import PolygonOptions


poly = PolygonOptions()

async def process_rsi(opts, ticker):
    try:
        rsi_snapshot= await opts.rsi_snapshot(ticker)
        print(rsi_snapshot)
        await opts.batch_insert_dataframe(rsi_snapshot, 'rsi', unique_columns='option_symbol, timespan')
    except Exception as e:
        print(e)

async def process_batch(opts, tickers):
    for ticker in tickers:
        await process_rsi(opts, ticker)
async def main():
    await poly.connect()
    tickers = await poly.get_tickers()

    batch_size = 250  # Process 250 tickers at a time
    tasks = []
    for i in range(0, len(tickers), batch_size):
        batch = tickers[i:i + batch_size]
        # Create an asynchronous task for each batch
        task = asyncio.create_task(process_batch(poly, batch))
        tasks.append(task)

    await asyncio.gather(*tasks)



asyncio.run(main())


import pandas as pd
class AvgInterestRates:
    def __init__(self, datas):
        self.record_date = [i.get('record_date') for i in datas]
        self.security_type_desc = [i.get('security_type_desc') for i in datas]
        self.security_desc = [i.get('security_desc') for i in datas]
        self.avg_interest_rate_amt = [i.get('avg_interest_rate_amt') for i in datas]
        self.src_line_nbr = [i.get('src_line_nbr') for i in datas]
        self.record_fiscal_year = [i.get('record_fiscal_year') for i in datas]
        self.record_fiscal_quarter = [i.get('record_fiscal_quarter') for i in datas]
        self.record_calendar_year = [i.get('record_calendar_year') for i in datas]
        self.record_calendar_quarter = [i.get('record_calendar_quarter') for i in datas]
        self.record_calendar_month = [i.get('record_calendar_month') for i in datas]
        self.record_calendar_day = [i.get('record_calendar_day') for i in datas]


        self.data_dict = { 

            'date': self.record_date,
            'security_type_description': self.security_type_desc,
            'security_description': self.security_desc,
            'avg_rate_amount': self.avg_interest_rate_amt,
            'source_line_number': self.src_line_nbr,
            'record_fiscal_year': self.record_fiscal_year,
            'record_fiscal_qtr': self.record_calendar_quarter,
            'record_calendar_year': self.record_calendar_year,
            'record_calendar_qtr': self.record_calendar_quarter,
            'record_calendar_month': self.record_calendar_month,
            'record_calendar_day': self.record_calendar_day
        }


        self.as_dataframe = pd.DataFrame(self.data_dict)


import os
import aiohttp
import asyncio
import pandas as pd
from dotenv import load_dotenv
from .models import Gex,GexMajorLevels,MaxGex


load_dotenv()

from typing import List

class GEXBot:
    def __init__(self):
        self.key = os.environ.get('GEXBOT')
        self.tickers = ["SPY","QQQ","AAPL","TSLA","MSFT","AMZN","NVDA"]
        self.df = None



    async def get_gex(self, ticker:str, as_dataframe:bool = True):
        """
        Gets GEX levels for a ticker.

        Arguments:


        >>> ticker: the ticker to survey (REQUIRED)
        
        >>> as_dataframe: returns as a dataframe (optional - default True)

        >>> concurrent: - optional - runs function for all tickers (default False)
        """
        url=f"https://api.gexbot.com/{ticker}/gex/all?key={self.key}"
        async with aiohttp.ClientSession() as session:
            async with session.get(url) as resp:
                data = await resp.json()
                if data is not None:
                    if as_dataframe == False:
                        return Gex(data)
                    data = Gex(data).as_dataframe
                   
                    return data
  
    async def major_levels(self, ticker, as_dataframe:bool=True):
        """
        Gets major gex levels

        Arguments:


        >>> ticker: required - the ticker to survey

        >>> as_dataframe: optional - returns as dataframe (default True)

        
        
        """
        url = f"https://api.gexbot.com/{ticker}/gex/all/majors?key={self.key}"
       
        async with aiohttp.ClientSession() as session:
            async with session.get(url) as resp:
                data = await resp.json()
                if data is not None:
                    if as_dataframe == False:
                        return GexMajorLevels(data)
                    
                    try:
                        # Attempt to use the as_dataframe attribute
                        dataframe = GexMajorLevels(data).as_dataframe
                        return dataframe
                    except AttributeError:
                        # Skip if the attribute does not exist
                        print(f'Error - {ticker}')
                
                

    async def max_gex(self,ticker, as_dataframe: bool=True):
        """Gets MAX gex levels
        
        Arguments:

        >>> ticker - required - the ticker to survey

        >>> as_dataframe - optional - returns as a dataframe (default True)
        

        """

        url = f"https://api.gexbot.com/{ticker}/gex/all/maxchange?key={self.key}"
        async with aiohttp.ClientSession() as session:
            async with session.get(url) as resp:
                data = await resp.json()
                if data is not None:
                    if as_dataframe == False:
                        return MaxGex(data)
                    return MaxGex(data).as_dataframe


    async def monitor_all_gex(self, ticker) -> List[Gex]:
        url = f"https://api.gexbot.com/{ticker}/gex/all?key={self.key}"
        all_data_dicts = []
        async with aiohttp.ClientSession() as session:
            async with session.get(url) as resp:
                data = await resp.json()
                if data is not None:
                    data = Gex(data)
                    all_data_dicts.append(data.data_dict)

        return all_data_dicts

    async def run_all_gex(self):
        tasks = [self.monitor_all_gex(i) for i in self.tickers]
        results = await asyncio.gather(*tasks)

        # Convert each list to a DataFrame before concatenation
        dataframes = [pd.DataFrame(r) for r in results if isinstance(r, list)]
        df = pd.concat(dataframes, ignore_index=True)
        df['timestamp'] = pd.to_datetime(df['timestamp'], unit='s')
        # Assuming df['timestamp'] is already in UTC
        df['timestamp'] = pd.to_datetime(df['timestamp']).dt.tz_localize('UTC').dt.tz_convert('US/Eastern')

        # Remove timezone information
        df['timestamp'] = df['timestamp'].dt.tz_localize(None)
        

        return df


import sys
from pathlib import Path




import os
import pandas as pd
from asyncio import Lock
lock = Lock()
from dotenv import load_dotenv
load_dotenv()
# Assuming we have the array of ticker IDs as numpy arrays from the user's environment
import numpy as np
from aiohttp.client_exceptions import ContentTypeError
from .options import WebullOptionsData, VolumeAnalysis
from .webull_trading import WebullTrading
from . import batch_insert_dataframe
import aiohttp
import asyncio
import json
import asyncpg
from ..helpers import get_human_readable_string
GEX_KEY = os.environ.get('GEXBOT')
def convert_to_date(date_str):
    try:
        return datetime.strptime(date_str, '%Y-%m-%d').date()
    except ValueError:
        # Handle the error or return None if the string cannot be converted
        return 
print(GEX_KEY)


from .webull_trading import WebullTrading
from datetime import datetime, timedelta

trading = WebullTrading()

class WebullOptions:
    def __init__(self, connection_string = None):
        self.pool = None
        self.most_active_tickers = ['SPY', 'QQQ', 'SPX', 'TSLA', 'AMZN', 'IWM', 'NVDA', 'VIX', 'AAPL', 'F', 'META', 'MSFT', 'GOOGL', 'HYG', 'INTC', 'SQQQ', 'AMD', 'TQQQ', 'XLF', 'BAC', 'XLI', 'TLT', 'GOOG', 'GLD', 'SOFI', 'EEM', 'EFA', 'UVXY', 'NFLX', 'ENPH', 'SQ', 'COIN', 'CVX', 'PLTR', 'XBI', 'FXI', 'XOM', 'VXX', 'PYPL', 'GDX', 'AAL', 'MARA', 'JPM', 'XLE', 'EWZ', 'PFE', 'BABA', 'AMC', 'SLV', 'SOXL', 'DIS', 'UBER', 'DIA', 'GM', 'CVNA', 'RIVN', 'RIOT', 'VALE', 'KRE', 'C', 'VZ', 'USO', 'BA', 'ARKK', 'X', 'MPW', 'XSP', 'NIO', 'SNAP', 'RUT', 'KVUE', 'EDR', 'SHOP', 'SMH', 'BMY', 'JNJ', 'KWEB', 'CHPT', 'MRNA', 'BITO', 'GOLD', 'ZM', 'T', 'NEM', 'ET', 'KO', 'PBR', 'MS', 'SCHW', 'OXY', 'MU', 'DKNG', 'RIG', 'MO', 'WFC', 'NDX', 'VFS', 'XLU', 'BKLN', 'MCD', 'ABBV', 'JBLU', 'FSLR', 'AI', 'LCID', 'SNOW', 'ABNB', 'TNA', 'DVN', 'DAL', 'RTX', 'JD', 'UNG', 'RBLX', 'TGT', 'ADBE', 'UPS', 'WDC', 'LUV', 'TSM', 'UAL', 'PAA', 'ORCL', 'PLUG', 'GS', 'LQD', 'CCL', 'LABU', 'EPD', 'WE', 'AFRM', 'XPO', 'MSOS', 'IBM', 'XLV', 'NKE', 'MSTR', 'COST', 'QCOM', 'HD', 'CSCO', 'AVGO', 'SPXS', 'CLF', 'TFC', 'GME', 'ON', 'CVS', 'CMG', 'SPXU', 'AGNC', 'XLY', 'COF', 'FCX', 'PDD', 'WMT', 'MTCH', 'NEE', 'XOP', 'CRM', 'ROKU', 'MA', 'RUN', 'SBUX', 'PARA', 'SE', 'V', 'SAVE', 'UPST', 'DXCM', 'LLY', 'NCLH', 'ABT', 'AXP', 'ABR', 'CHWY', 'AA', 'DDOG', 'SVXY', 'LYFT', 'RCL', 'HOOD', 'BEKE', 'IBB', 'LI', 'PINS', 'PANW', 'ETSY', 'YINN', 'SAVA', 'OIH', 'WBA', 'TXN', 'FEZ', 'PG', 'CCJ', 'BOIL', 'SMCI', 'ALGN', 'XLP', 'CRWD', 'GE', 'MRVL', 'BX', 'WBD', 'SOXS', 'MRK', 'W', 'UVIX', 'SPXL', 'FSR', 'TZA', 'URNM', 'CAT', 'PEP', 'IMGN', 'XPEV', 'LULU', 'CVE', 'TTD', 'CMCSA', 'BIDU', 'NLY', 'AX', 'XRT', 'AG', 'BYND', 'BRK B', 'HL', 'M', 'NWL', 'SEDG', 'SIRI', 'EBAY', 'FLEX', 'BTU', 'NKLA', 'DISH', 'MDT', 'PSEC', 'VMW', 'ZS', 'COP', 'DG', 'AMAT', 'UCO', 'MDB', 'SLB', 'PTON', 'OKTA', 'U', 'HSBC', 'XHB', 'TMUS', 'UNH', 'OSTK', 'CGC', 'NOW', 'TLRY', 'DOCU', 'TDOC', 'MMM', 'HPQ', 'PCG', 'CHTR', 'Z', 'LOW', 'PENN', 'LMT', 'WOLF', 'KMI', 'VLO', 'SPWR', 'XLK', 'DLTR', 'WHR', 'NVAX', 'ARM', 'JETS', 'VNQ', 'DE', 'DLR', 'NET', 'FAS', 'WPM', 'DASH', 'ACN', 'ASHR', 'FUBO', 'CLX', 'ADM', 'SRPT', 'MRO', 'KGC', 'DPST', 'TWLO', 'AR', 'CNC', 'FDX', 'AMGN', 'VRT', 'CLSK', 'EMB', 'KOLD', 'CD', 'HES', 'SPOT', 'XLC', 'ZIM', 'GILD', 'EQT', 'CRSP', 'GDXJ', 'STNG', 'NAT', 'HAL', 'SGEN', 'GPS', 'USB', 'QS', 'UPRO', 'KSS', 'IDXX', 'FTNT', 'BALL', 'TMF', 'PACW', 'EL', 'MULN', 'NVO', 'GDDY', 'BBY', 'SPCE', 'SNY', 'KEY', 'MGM', 'FREY', 'CZR', 'LVS', 'TTWO', 'LRCX', 'MXEF', 'PAGP', 'ANET', 'VFC', 'GRPN', 'EW', 'BKNG', 'EOSE', 'TMO', 'SPY', 'SPX', 'QQQ', 'VIX', 'IWM', 'TSLA', 'HYG', 'AMZN', 'AAPL', 'BAC', 'XLF', 'TLT', 'SLV', 'EEM', 'F', 'NVDA', 'GOOGL', 'AMD', 'AAL', 'META', 'INTC', 'PLTR', 'C', 'GLD', 'MSFT', 'GDX', 'FXI', 'VALE', 'GOOG', 'XLE', 'SOFI', 'BABA', 'NIO', 'PFE', 'EWZ', 'PYPL', 'T', 'CCL', 'SNAP', 'DIS', 'GM', 'NKLA', 'WFC', 'TQQQ', 'AMC', 'UBER', 'RIVN', 'KRE', 'PBR', 'XOM', 'LCID', 'MARA', 'JPM', 'GOLD', 'ET', 'PLUG', 'JD', 'VZ', 'WBD', 'EFA', 'KVUE', 'RIG', 'SQ', 'CHPT', 'KWEB', 'KO', 'MU', 'BITO', 'TSM', 'SQQQ', 'SHOP', 'DKNG', 'CSCO', 'XLU', 'COIN', 'MPW', 'OXY', 'SOXL', 'FCX', 'RIOT', 'DAL', 'SCHW', 'TLRY', 'BA', 'NFLX', 'UAL', 'SIRI', 'MS', 'AGNC', 'UVXY', 'XBI', 'PARA', 'ARKK', 'CMCSA', 'DVN', 'UNG', 'VXX', 'CVX', 'CLF', 'RBLX', 'PINS', 'XLI', 'SE', 'CVNA', 'QCOM', 'SGEN', 'USO', 'TMF', 'BMY', 'RTX', 'XSP', 'ORCL', 'WBA', 'NKE', 'PDD', 'X', 'KMI', 'GME', 'NCLH', 'NEM', 'SMH', 'MSOS', 'TEVA', 'M', 'XPEV', 'ABBV', 'JETS', 'ABNB', 'MULN', 'JNJ', 'MO', 'CVS', 'AFRM', 'LUV', 'NEE', 'FSR', 'AI', 'SAVE', 'JBLU', 'HOOD', 'ENPH', 'DIA', 'WMT', 'LYFT', 'NU', 'BP', 'XOP', 'ENVX', 'SPCE', 'NOK', 'GRAB', 'BYND', 'ZM', 'SLB', 'NVAX', 'U', 'MRVL', 'CCJ', 'OPEN', 'CRM', 'CGC', 'AA', 'V', 'IBM', 'PTON', 'SBUX', 'LABU', 'TGT', 'STNE', 'BRK B', 'ASHR', 'UPST', 'QS', 'MRK', 'MRNA', 'VFS', 'XHB', 'TMUS', 'SNOW', 'PANW', 'VFC', 'UPS', 'BX', 'DISH', 'USB', 'TFC', 'GE', 'COP', 'LI', 'MET', 'XRT', 'ROKU', 'XLP', 'CHWY', 'FSLR', 'PG', 'XLK', 'FUBO', 'XLV', 'W', 'AMAT', 'GOEV', 'TXN', 'PEP', 'RUN', 'SWN', 'DOW', 'HD', 'GS', 'KGC', 'Z', 'AG', 'ABR', 'CAT', 'UUP', 'AXP', 'ZIM', 'KHC', 'RCL', 'LAZR', 'BOIL', 'DDOG', 'PENN', 'TTD', 'TELL', 'XLY', 'EPD', 'CRWD', 'VMW', 'NYCB', 'HUT', 'BTU', 'DOCU', 'NET', 'BKLN', 'SU', 'BAX', 'ETSY', 'HE', 'BTG', 'NLY', 'BHC', 'TDOC', 'LUMN', 'CLSK', 'MCD', 'LVS', 'MMM', 'DM', 'ALLY', 'SPWR', 'VRT', 'ABT', 'DASH', 'ADBE', 'TNA', 'MA', 'ACB', 'MDT', 'MGM', 'COST', 'WDC', 'GSAT', 'GPS', 'ON', 'MRO', 'PAAS', 'EOSE', 'LQD', 'BILI', 'AR', 'ONON', 'HTZ', 'TWLO', 'GILD', 'MMAT', 'ASTS', 'STLA', 'LLY', 'SABR', 'BIDU', 'EDR', 'AVGO', 'HAL', 'DG', 'WYNN', 'AEM', 'PATH', 'DB', 'IYR', 'UNH', 'HL', 'IEF', 'SPXS', 'CPNG', 'URA', 'NVO', 'BITF', 'URNM', 'KSS', 'FTCH', 'KEY', 'TH', 'GEO', 'FDX', 'CL', 'AZN', 'HPQ', 'DNN', 'BSX', 'SHEL', 'DXCM', 'PCG', 'BEKE', 'DNA', 'PM', 'TTWO', 'IQ', 'WE', 'ALB', 'SAVA', 'GDXJ', 'SPXU', 'OSTK', 'COF', 'SNDL', 'OKTA', 'BXMT', 'UEC', 'VLO', 'KR', 'ZION', 'WW', 'RSP', 'XP', 'IAU', 'LULU', 'ARCC', 'SOXS', 'VOD', 'TJX', 'MOS', 'EQT', 'IONQ', 'STNG', 'NOVA', 'HLF', 'HSBC', 'ARM']
        self.most_active_tickers = set(self.most_active_tickers)
        self.gex_tickers = ['SPY','SPX','QQQ','AAPL','TSLA','MSFT','AMZN','NVDA']
        self.as_dataframe = None
        self.db_url = connection_string
        self.connection_string = connection_string
        self.api_key = os.environ.get('YOUR_POLYGON_KEY')
        self.today = datetime.now().strftime('%Y-%m-%d')
        self.yesterday = (datetime.now() - timedelta(days=1)).strftime('%Y-%m-%d')
        self.tomorrow = (datetime.now() + timedelta(days=1)).strftime('%Y-%m-%d')
        self.thirty_days_ago = (datetime.now() - timedelta(days=30)).strftime('%Y-%m-%d')
        self.thirty_days_from_now = (datetime.now() + timedelta(days=30)).strftime('%Y-%m-%d')
        self.fifteen_days_ago = (datetime.now() - timedelta(days=15)).strftime('%Y-%m-%d')
        self.fifteen_days_from_now = (datetime.now() + timedelta(days=15)).strftime('%Y-%m-%d')
        self.eight_days_from_now = (datetime.now() + timedelta(days=8)).strftime('%Y-%m-%d')
        self.eight_days_ago = (datetime.now() - timedelta(days=8)).strftime('%Y-%m-%d')
        self.headers = {
        "Access_token": os.environ.get('ACCESS_TOKEN'),
        "Accept": "*/*",
        "App": "global",
        "App-Group": "broker",
        "Appid": "wb_web_app",
        "Content-Type": "application/json;charset=UTF-8",
        "Device-Type": "Web",
        "Did": os.environ.get('DID'),
        "Hl": "en",
        "Locale": "eng",
        "Os": "web",
        "Osv": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/119.0.0.0 Safari/537.36",
        "Ph": "Windows Chrome",
        "Platform": "web",
        "Referer": "https://app.webull.com/",
        "Sec-Ch-Ua": "\"Chromium\";v=\"118\", \"Google Chrome\";v=\"118\", \"Not=A?Brand\";v=\"99\"",
        "Sec-Ch-Ua-Mobile": "?0",
        "Sec-Ch-Ua-Platform": "\"Windows\"",
        "T_time": "1698276695206",
        "Tz": "America/Los_Angeles",
        "User-Agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/118.0.0.0 Safari/537.36",
    }


    async def connect(self):
        self.pool = await asyncpg.create_pool(
            dsn=self.connection_string, min_size=1, max_size=100
        )

    async def create_table(self, df, table_name):
        print("Connected to the database.")
        dtype_mapping = {
            'int64': 'BIGINT',
            'float64': 'DOUBLE PRECISION',
            'object': 'TEXT',
            'bool': 'BOOLEAN',
            'datetime.date': 'TIMESTAMP',
            'datetime.datetime': 'TIMESTAMP',
            'datetime64[ns]': 'timestamp',
            'datetime64[ms]': 'timestamp',
            'datetime64[ns, US/Eastern]': 'TIMESTAMP WITH TIME ZONE',
            'string': 'TEXT'
        }


        # Check for large integers and update dtype_mapping accordingly
        for col, dtype in zip(df.columns, df.dtypes):
            if dtype == 'int64':
                max_val = df[col].max()
                min_val = df[col].min()
                if max_val > 2**31 - 1 or min_val < -2**31:
                    dtype_mapping['int64'] = 'BIGINT'
        history_table_name = f"{table_name}_history"
        async with self.pool.acquire() as connection:

            table_exists = await connection.fetchval(f"SELECT to_regclass('{table_name}')")
            
            if table_exists is None:
                create_query = f"""
                CREATE TABLE {table_name} (
                    {', '.join(f'"{col}" {dtype_mapping[str(dtype)]}' for col, dtype in zip(df.columns, df.dtypes))}
                )
                """

                print(f"Creating table with query: {create_query}")

                # Create the history table
                history_create_query = f"""
                CREATE TABLE IF NOT EXISTS {history_table_name} (
                    id serial PRIMARY KEY,
                    operation CHAR(1) NOT NULL,
                    changed_at TIMESTAMP NOT NULL DEFAULT current_timestamp,
                    {', '.join(f'"{col}" {dtype_mapping[str(dtype)]}' for col, dtype in zip(df.columns, df.dtypes))}
                );
                """
                print(f"Creating history table with query: {history_create_query}")
                await connection.execute(history_create_query)
                try:
                    await connection.execute(create_query)
                    print(f"Table {table_name} created successfully.")
                except asyncpg.UniqueViolationError as e:
                    print(f"Unique violation error: {e}")
            else:
                print(f"Table {table_name} already exists.")
            
            # Create the trigger function
            trigger_function_query = f"""
            CREATE OR REPLACE FUNCTION save_to_{history_table_name}()
            RETURNS TRIGGER AS $$
            BEGIN
                INSERT INTO {history_table_name} (operation, changed_at, {', '.join(f'"{col}"' for col in df.columns)})
                VALUES (
                    CASE
                        WHEN (TG_OP = 'DELETE') THEN 'D'
                        WHEN (TG_OP = 'UPDATE') THEN 'U'
                        ELSE 'I'
                    END,
                    current_timestamp,
                    {', '.join('OLD.' + f'"{col}"' for col in df.columns)}
                );
                RETURN NEW;
            END;
            $$ LANGUAGE plpgsql;
            """
            await connection.execute(trigger_function_query)

            # Create the trigger
            trigger_query = f"""
            DROP TRIGGER IF EXISTS tr_{history_table_name} ON {table_name};
            CREATE TRIGGER tr_{history_table_name}
            AFTER UPDATE OR DELETE ON {table_name}
            FOR EACH ROW EXECUTE FUNCTION save_to_{history_table_name}();
            """
            await connection.execute(trigger_query)


            # Alter existing table to add any missing columns
            for col, dtype in zip(df.columns, df.dtypes):
                alter_query = f"""
                DO $$
                BEGIN
                    BEGIN
                        ALTER TABLE {table_name} ADD COLUMN "{col}" {dtype_mapping[str(dtype)]};
                    EXCEPTION
                        WHEN duplicate_column THEN
                        NULL;
                    END;
                END $$;
                """
                await connection.execute(alter_query)

    async def fetch(self, query, *args):
        """
        Fetch data from the database using the provided SQL query.

        :param query: The SQL query to execute.
        :param args: The arguments to pass to the SQL query.
        :return: The result of the query as a list of records.
        """
        async with self.pool.acquire() as conn:  # Acquire a connection from the pool
            # Execute the query with the provided arguments
            return await conn.fetch(query, *args)


    async def batch_insert_options(self, pairs):
        try:
            await self.connect()
            async with self.pool.acquire() as conn:  # Acquire a connection from the pool

                async with conn.transaction():  # Start a transaction
                    # Prepare the statement to insert data
                    insert_query = 'INSERT INTO webull_options (symbol, ticker_id, ticker) VALUES ($1, $2, $3)'
                    # Perform the batch insert
                    await conn.executemany(insert_query, pairs)
                    print("Batch insert completed.")
        except asyncpg.exceptions.UniqueViolationError:
            print(f'Duplicate found - skipping.')


    async def yield_batch_ids(self, ticker_symbol):
   
        async with self.pool.acquire() as conn:
            # We will fetch all derivative IDs associated with the ticker symbol
            derivative_ids = await conn.fetch(
                'SELECT ticker_id FROM webull_options WHERE ticker = $1',
                ticker_symbol
            )
            
            # Convert the records to a list of IDs
            derivative_id_list = [str(record['ticker_id']) for record in derivative_ids]

            # Yield batches of 55 IDs at a time as a comma-separated string
            for i in range(0, len(derivative_id_list), 55):
                yield ','.join(derivative_id_list[i:i+55])

    async def get_option_ids(self, ticker):
        ticker_id = await trading.get_ticker_id(ticker)
        params = {
            "tickerId": f"{ticker_id}",
            "count": -1,
            "direction": "all",
            "expireCycle": [1,
                3,
                2,
                4
            ],
            "type": 0,
            "quoteMultiplier": 100,
            "unSymbol": f"{ticker}"
        }
        data = json.dumps(params)
        url="https://quotes-gw.webullfintech.com/api/quote/option/strategy/list"

        # Headers you may need to include, like authentication tokens, etc.
        headers = trading.headers
        # The body of your POST request as a Python dictionary
        import pandas as pd
        # Make the POST request
        # Make the POST request
        async with aiohttp.ClientSession(headers=headers) as session:
            async with session.post(url, data=data) as resp:
                response_json = await resp.json()
        
                # Extract the 'expireDateList' from the response
                expireDateList = response_json.get('expireDateList')

                # Flatten the nested 'data' from each item in 'expireDateList'
            try:
                data_flat = [item for sublist in expireDateList if sublist and sublist.get('data') for item in sublist['data']]



                # Create a DataFrame from the flattened data
                df_cleaned = pd.DataFrame(data_flat)

                # Drop the 'askList' and 'bidList' columns if they exist
                df_cleaned.drop(columns=['askList', 'bidList'], errors='ignore', inplace=True)
                # Existing DataFrame columns
                df_columns = df_cleaned.columns

                # Original list of columns you want to convert to numeric
                numeric_cols = ['open', 'high', 'low', 'strikePrice', 'isStdSettle', 'quoteMultiplier', 'quoteLotSize']

                # Filter the list to include only columns that exist in the DataFrame
                existing_numeric_cols = [col for col in numeric_cols if col in df_columns]

                # Now apply the to_numeric conversion only to the existing columns
                df_cleaned[existing_numeric_cols] = df_cleaned[existing_numeric_cols].apply(pd.to_numeric, errors='coerce')

      
            
                df_cleaned.to_csv('test.csv', index=False)


                # Load the data from the CSV file
                df = pd.read_csv('test.csv')

                # Extract 'tickerId' column values in batches of 55
                ticker_ids = df['tickerId'].unique()  # Assuming 'tickerId' is a column in your DataFrame
                symbol_list = df['symbol'].unique().tolist()
            # Pair up 'tickerId' and 'symbol'
                # Before you call batch_insert_options, make sure pairs contain the correct types
                pairs = [(str(symbol), int(ticker_id), str(ticker)) for ticker_id, symbol in zip(ticker_ids, symbol_list)]

                
                await self.batch_insert_options(pairs)
                return pairs
            except (ContentTypeError, TypeError):
                print(f'Error for {ticker}')
    async def get_option_id_for_symbol(self, ticker_symbol):
        async with self.pool.acquire() as conn:
            # Start a transaction
            async with conn.transaction():
                # Execute the query to get the option_id for a given ticker_symbol
                # This assumes 'symbol' column exists in 'options_data' table and 
                # is used to store the ticker symbol
                query = f'''
                    SELECT ticker_id FROM wb_opts
                    WHERE underlying_symbol = '{ticker_symbol}';
                '''
                # Fetch the result
                result = await conn.fetch(query)
                # Return a list of option_ids or an empty list if none were found
                return [record['ticker_id'] for record in result]


    async def get_option_symbols_by_ticker_id(self, ticker_id):
        async with self.pool.acquire() as conn:
            # Start a transaction
            async with conn.transaction():
                # Execute the query to get all option_symbols for a given ticker_id
                query = '''
                    SELECT option_symbol FROM options_data
                    WHERE ticker_id = $1;
                '''
                # Fetch the result
                records = await conn.fetch(query, ticker_id)
                # Extract option_symbols from the records
                return [record['option_symbol'] for record in records]
    async def get_ticker_symbol_pairs(self):
        # Assume 'pool' is an instance variable pointing to a connection pool
        async with self.pool.acquire() as conn:
            # Start a transaction
            async with conn.transaction():
                # Create a cursor for iteration using 'cursor()' instead of 'execute()'
                async for record in conn.cursor('SELECT ticker_id, symbol FROM webull_options'):
                    yield (record['ticker_id'], record['symbol'])

    async def get_volume_analysis(self, ticker):
        ticker_id = await trading.get_ticker_id(ticker)
        params = {
            "tickerId": f"{ticker_id}",
            "count": -1,
            "direction": "all",
            "expireCycle": [1,
                3,
                2,
                4
            ],
            "type": 0,
            "quoteMultiplier": 100,
            "unSymbol": f"{ticker}"
        }
        data = json.dumps(params)
        url="https://quotes-gw.webullfintech.com/api/quote/option/strategy/list"

        # Headers you may need to include, like authentication tokens, etc.
        headers = trading.headers
        # The body of your POST request as a Python dictionary
        import pandas as pd
        # Make the POST request
        # Make the POST request
        async with aiohttp.ClientSession(headers=headers) as session:
            async with session.post(url, data=data) as resp:
                response_json = await resp.json()
          
                # Extract the 'expireDateList' from the response
                expireDateList = response_json.get('expireDateList')

                # Flatten the nested 'data' from each item in 'expireDateList'
            try:
                data_flat = [item for sublist in expireDateList if sublist and sublist.get('data') for item in sublist['data']]



                # Create a DataFrame from the flattened data
                df_cleaned = pd.DataFrame(data_flat)

                # Drop the 'askList' and 'bidList' columns if they exist
                df_cleaned.drop(columns=['askList', 'bidList'], errors='ignore', inplace=True)

                # Convert specified columns to numeric values, coercing errors to NaN
                numeric_cols = ['open', 'high', 'low', 'strikePrice', 'isStdSettle', 'quoteMultiplier', 'quoteLotSize']
                # Iterate through the list of numeric columns and check if they exist in df_cleaned
                existing_numeric_cols = [col for col in numeric_cols if col in df_cleaned.columns]

                # Now apply the conversion only on the columns that exist
                df_cleaned[existing_numeric_cols] = df_cleaned[existing_numeric_cols].apply(pd.to_numeric, errors='coerce')

                print(df_cleaned)
                df_cleaned.to_csv('test.csv', index=False)


                # Load the data from the CSV file
                df = pd.read_csv('test.csv')

                # Extract 'tickerId' column values in batches of 55
                ticker_ids = df['tickerId'].unique()  # Assuming 'tickerId' is a column in your DataFrame
                symbol_list = df['symbol'].unique().tolist()
            # Pair up 'tickerId' and 'symbol'
                pairs = list(zip(ticker_ids, symbol_list))

                
                # Split into batches of 55
                batches = [ticker_ids[i:i + 55] for i in range(0, len(ticker_ids), 55)]

                ticker_id_strings = [','.join(map(str, batch)) for batch in batches]







                for ticker_id_string in ticker_id_strings:
                    ticker_ids = ticker_id_string.split(',')
                    for deriv_id in ticker_ids:
                        all_data = []
                        volume_analysis_url = f"https://quotes-gw.webullfintech.com/api/statistic/option/queryVolumeAnalysis?count=200&tickerId={deriv_id}"
                        async with aiohttp.ClientSession(headers=headers) as session:
                            async with session.get(volume_analysis_url) as resp:
                                data = await resp.json()
                                all_data.append(data)


                   
                        return all_data
                        #df = pd.DataFrame(all_data)
                        #df.to_csv('all_options', index=False)
            except (ContentTypeError, TypeError):
                print(f'Error for {ticker}')

    async def get_option_ids_limited(self, sem, ticker):
        async with sem:
            # This will wait until the semaphore allows entry (i.e., under the limit)
            return await self.get_option_ids(ticker)

    async def harvest_options(self,most_active_tickers):
        # Set the maximum number of concurrent requests
        max_concurrent_requests = 5  # For example, limiting to 10 concurrent requests

        # Create a semaphore with your desired number of concurrent requests
        sem = asyncio.Semaphore(max_concurrent_requests)
        await self.connect()
        # Create tasks using the semaphore
        tasks = [self.get_option_ids_limited(sem, ticker) for ticker in most_active_tickers]

        # Run the tasks concurrently and wait for all to complete
        await asyncio.gather(*tasks)


    async def get_option_data_for_ticker(self, ticker):
        print(f"Starting processing for ticker: {ticker}")
        dataframes = []  # Initialize a list to collect DataFrames
  
        async for info in self.yield_batch_ids(ticker_symbol=ticker):
            print(f"Processing batch ID: {info} for ticker: {ticker}")
            url = f"https://quotes-gw.webullfintech.com/api/quote/option/quotes/queryBatch?derivativeIds={info}"
            async with aiohttp.ClientSession(headers=trading.headers) as session:
                async with session.get(url) as resp:
                    data = await resp.json()
                    if not data:  # If data is empty or None, break the loop
                        print(f"No more data for ticker: {ticker}. Moving to next.")
                        break
                    wb_data = WebullOptionsData(data)
                    if self.as_dataframe is not None:
                        df = wb_data.as_dataframe
                        df['ticker'] = ticker
                        df = df.rename(columns={'open_interest_change': 'oi_change'})
                        
                        await self.insert_dataframe_in_batches(df, 'options_data')
                        dataframes.append(df)
        return dataframes
            
        

  



    async def insert_dataframe_in_batches(self, df, table_name, batch_size=55):
        """
        Insert a pandas DataFrame into a SQL table in batches.

        :param df: The pandas DataFrame to insert.
        :param table_name: The name of the target SQL table.
        :param batch_size: The size of the batches to insert.
        """
        # Make sure we have a connection
        
        df['expiry_date'] = df['expiry_date'].apply(convert_to_date)
        df['close'] = df['close'].astype(float)
        df['open'] = df['open'].astype(float)
        df['high'] = df['high'].astype(float)
        df['low'] = df['low'].astype(float)
        df['open_interest'] = df['open_interest'].replace({np.nan: None})
        async with self.pool.acquire() as conn:
            try:
                # Convert DataFrame to list of tuples
                records = df.to_records(index=False)
                columns = df.columns.tolist()
                values = [tuple(x) for x in records]

                # Create a prepared statement
                placeholders = ', '.join(f'${i+1}' for i in range(len(columns)))
                insert_query = f'INSERT INTO {table_name} ({", ".join(columns)}) VALUES ({placeholders})'

                # Insert data in batches
                for i in range(0, len(values), batch_size):
                    batch = values[i:i + batch_size]
                    await conn.executemany(insert_query, batch)

            except asyncpg.exceptions.UniqueViolationError:
                print(f'Duplicate found - skipping.')


        # Initialize an HTTP session
    async def associate_dates_with_data(self, dates, datas):
        if datas is not None and dates is not None:
        # This function remains for your specific data handling if needed
            return [{**data, 'date': date} for date, data in zip(dates, datas)]
    async def fetch_volume_analysis(self, option_symbol, id, underlying_ticker):
        url = f"https://quotes-gw.webullfintech.com/api/statistic/option/queryVolumeAnalysis?count=200&tickerId={id}"
        async with aiohttp.ClientSession(headers=self.headers) as session:
            async with session.get(url) as resp:
                if resp.status == 200:
                    vol_anal = await resp.json()
                    dates = vol_anal.get('dates')
                    datas = vol_anal.get('datas')
                    associated_data = await self.associate_dates_with_data(dates, datas)

                    df = pd.DataFrame(associated_data)
                    df['option_symbol'] = option_symbol
                    components = get_human_readable_string(option_symbol)
                    df['underlying_ticker'] = underlying_ticker
                    df['strike'] = components.get('strike_price')
                    df['call_put'] = components.get('call_put')
                    df['expiry'] = components.get('expiry_date')
                    return df
                else:
                    print(f"Failed to fetch data for ID {id}: HTTP Status {resp.status}")
                    return pd.DataFrame()
                    
    # Initialize an HTTP session
    async def test_vol_anal(self, ticker_id):
        volume_analysis_url = f"https://quotes-gw.webullfintech.com/api/statistic/option/queryVolumeAnalysis?count=500&tickerId={ticker_id}"
        async with aiohttp.ClientSession(headers=self.headers) as session:
            async with session.get(volume_analysis_url) as resp:
                data = await resp.json()
                if data is not None:
                    return VolumeAnalysis(data).trades_and_dates_dict
    async def insert_trades_and_dates(self, data):
        data['date'] = datetime.strptime(data['date'], '%Y-%m-%d')
        # Convert numeric strings to appropriate numeric types
        data['price'] = float(data['price']) if data['price'] else 0.0
        data['buy'] = int(data['buy']) if data['buy'] else 0
        data['sell'] = int(data['sell']) if data['sell'] else 0
        data['volume'] = int(data['volume']) if data['volume'] else 0
        data['strike_price'] = float(data['strike_price']) if data['strike_price'] else 0.0
        # Convert expiry_date string to date object
        data['expiry_date'] = datetime.strptime(data['expiry_date'], '%Y-%m-%d')

                

        # Assuming 'data' is a dictionary containing the keys that match your table's columns.
        async with self.pool.acquire() as connection:
            await connection.execute('''
                INSERT INTO trades_and_dates (
                    date, price, buy, sell, volume, ratio,
                    option_symbol, symbol, strike_price, call_put,
                    expiry_date
                ) VALUES($1, $2, $3, $4, $5, $6, $7, $8, $9, $10, $11)
            ''',
            data['date'], data['price'], data['buy'], data['sell'], data['volume'],
            data['ratio'], data['option_symbol'], data['symbol'], data['strike_price'],
            data['call_put'], data['expiry_date'])
    async def insert_volume_analysis_data(self, data: dict):
        # Assuming data contains the fields corresponding to your self.data_dict
        # You would extract these values and insert them into your volume_analysis table
        async with self.pool.acquire() as conn:
            await conn.execute("""
                INSERT INTO volume_analysis (ticker_id, symbol, option_id, total_trades, total_volume, avg_price, buy_volume, sell_volume, neutral_volume, option_symbol, strike_price, call_put, expiry_date)
                VALUES ($1, $2, $3, $4, $5, $6, $7, $8, $9, $10, $11, $12, $13)
            """, data.get('ticker_id'), data.get('symbol', None), data.get('option_id', None), data.get('total_trades', None), data.get('total_volume', None), data.get('avg_price',None),data.get('buy_volume',None), data.get('sell_volume', None), data.get('neutral_volume', None), data.get('option_symbol', None), data.get('strike_price') , data.get('call_put'), data.get('expiry_date')        )
        
    async def run_all_tasks(self, tickers):

        async with aiohttp.ClientSession(headers=self.headers) as session:
            tasks = [self.get_option_data_for_ticker(ticker, session) for ticker in tickers]
            await asyncio.gather(*tasks)
    async def store_options_data(self):
        await self.connect()
        semaphore = asyncio.Semaphore(5)  # Adjust the number to limit concurrent tasks

        async def limited_get_option_data_for_ticker(ticker):
            async with semaphore:
                return await self.get_option_data_for_ticker(ticker)

        tasks = [limited_get_option_data_for_ticker(i) for i in self.most_active_tickers]
        await asyncio.gather(*tasks)

    async def get_vol1y(self, ticker):
        await self.connect()
        data = await self.get_option_id_for_symbol(ticker)
        print(data)
        for id in data:
            ticker_id = await WebullTrading().get_ticker_id(ticker)
            url = f"https://quotes-gw.webullfintech.com/api/quote/option/quotes/detail?derivativeIds={id}&tickerId={ticker_id}"
            async with aiohttp.ClientSession(headers=self.headers) as session:
                async with session.get(url) as resp:
                    data = await resp.json()
                    
                    vol1y = data.get('vol1y')
                    return vol1y
    async def batch_insert_dataframe(self, df, table_name, unique_columns, batch_size=250):
        """
        WORKS - Creates table - inserts data based on DTYPES.
        
        """
    
        async with lock:
            if not await self.table_exists(table_name):
                await self.create_table(df, table_name)
            
            # Debug: Print DataFrame columns before modifications
            #print("Initial DataFrame columns:", df.columns.tolist())
            
            df = df.copy()
            df.dropna(inplace=True)
            df['insertion_timestamp'] = [datetime.now() for _ in range(len(df))]

            # Debug: Print DataFrame columns after modifications
            #print("Modified DataFrame columns:", df.columns.tolist())
            
            records = df.to_records(index=False)
            data = list(records)


            async with self.pool.acquire() as connection:
                column_types = await connection.fetch(
                    f"SELECT column_name, data_type FROM information_schema.columns WHERE table_name = '{table_name}'"
                )
                type_mapping = {col: next((item['data_type'] for item in column_types if item['column_name'] == col), None) for col in df.columns}

                async with connection.transaction():
                    insert_query = f"""
                    INSERT INTO {table_name} ({', '.join(f'"{col}"' for col in df.columns)}) 
                    VALUES ({', '.join('$' + str(i) for i in range(1, len(df.columns) + 1))})
                    ON CONFLICT ({unique_columns})
                    DO UPDATE SET {', '.join(f'"{col}" = excluded."{col}"' for col in df.columns)}
                    """
            
                    batch_data = []
                    for record in data:
                        new_record = []
                        for col, val in zip(df.columns, record):
                
                            pg_type = type_mapping[col]

                            if val is None:
                                new_record.append(None)
                            elif pg_type == 'timestamp' and isinstance(val, np.datetime64):
                                new_record.append(pd.Timestamp(val).to_pydatetime().replace(tzinfo=None))

            
                            elif isinstance(val, datetime):
                                new_record.append(pd.Timestamp(val).to_pydatetime())
                            elif pg_type in ['timestamp', 'timestamp without time zone', 'timestamp with time zone'] and isinstance(val, np.datetime64):
                                new_record.append(pd.Timestamp(val).to_pydatetime().replace(tzinfo=None))  # Modified line
                            elif pg_type in ['double precision', 'real'] and not isinstance(val, str):
                                new_record.append(float(val))
                            elif isinstance(val, np.int64):  # Add this line to handle numpy.int64
                                new_record.append(int(val))
                            elif pg_type == 'integer' and not isinstance(val, int):
                                new_record.append(int(val))
                            else:
                                new_record.append(val)
                    
                        batch_data.append(new_record)

                        if len(batch_data) == batch_size:
                            try:
                                
                            
                                await connection.executemany(insert_query, batch_data)
                                batch_data.clear()
                            except Exception as e:
                                print(f"An error occurred while inserting the record: {e}")
                                await connection.execute('ROLLBACK')
                                raise

                if batch_data:  # Don't forget the last batch
    
                    try:

                        await connection.executemany(insert_query, batch_data)
                    except Exception as e:
                        print(f"An error occurred while inserting the record: {e}")
                        await connection.execute('ROLLBACK')
                        raise
    async def save_to_history(self, df, main_table_name, history_table_name):
        # Assume the DataFrame `df` contains the records to be archived
        if not await self.table_exists(history_table_name):
            await self.create_table(df, history_table_name, None)

        df['archived_at'] = datetime.now()  # Add an 'archived_at' timestamp
        await self.batch_insert_dataframe(df, history_table_name, None)
    async def table_exists(self, table_name):
        query = f"SELECT EXISTS (SELECT FROM information_schema.tables WHERE table_name = '{table_name}');"

        async with self.pool.acquire() as conn:
            async with conn.transaction():
                exists = await conn.fetchval(query)
        return exists

    async def close_pool(self):
        await self.pool.close()



from dataclasses import dataclass, asdict
from datetime import datetime

import pandas as pd

@dataclass
class FailureToDeliverData:
    date: datetime
    ticker: str
    ftd: int
    price: float
    ftd_x_dollar: float
    t_35_date: datetime

    def as_dataframe(self) -> pd.DataFrame:
        """Converts the object to a pandas DataFrame."""
        return pd.DataFrame([asdict(self)])

    def insert_into_pg(self, conn):
        cursor = conn.cursor()
        insert_query = """INSERT INTO failure_to_deliver (date, ticker, ftd, price, ftd_x_dollar, t_35_date) VALUES (%s, %s, %s, %s, %s, %s);"""
        cursor.execute(insert_query, (self.date, self.ticker, self.ftd, self.price, self.ftd_x_dollar, self.t_35_date))
        conn.commit()



import pandas as pd


class TickerInfo:
    def __init__(self, data):
        self.tickerId = [i['tickerId'] for i in data]
        self.exchangeId = [i['exchangeId'] for i in data]
        self.regionId = [i['regionId'] for i in data]
        self.symbol = [i['symbol'] for i in data]
        self.unSymbol = [i['unSymbol'] for i in data]
        self.tickerType = [i['tickerType'] for i in data]
        self.belongTickerId = [i['belongTickerId'] for i in data]
        self.direction = [i['direction'] for i in data]
        self.quoteLotSize = [i['quoteLotSize'] for i in data]
        self.expireDate = [i['expireDate'] for i in data]
        self.strikePrice = [i['strikePrice'] for i in data]
        self.change = [i['change'] for i in data]
        self.changeRatio = [i['changeRatio'] for i in data]
        self.quoteMultiplier = [i['quoteMultiplier'] for i in data]
        self.cycle = [i['cycle'] for i in data]

        self.data_dict = { 
            'tickerId': self.tickerId,
            'exchangeId': self.exchangeId,
            'regionId': self.regionId,
            'symbol': self.symbol,
            'unSymbol': self.unSymbol,
            'tickerType': self.tickerType,
            'belongTickerId': self.belongTickerId,
            'direction': self.direction,
            'quoteLotSize': self.quoteLotSize,
            'expireDate': self.expireDate,
            'strikePrice': self.strikePrice,
            'change': self.change,
            'changeRatio': self.changeRatio,
            'quoteMultiplier': self.quoteMultiplier,
            'cycle': self.cycle
        }

        self.as_dataframe = pd.DataFrame(self.data_dict)



class ScreenerRule:
    def __init__(self, data):
        self.change = [i['options.screener.rule.change'] for i in data]
        self.expireDate = [i['options.screener.rule.expireDate'] for i in data]
        self.ask = [i['options.screener.rule.ask'] for i in data]
        self.openInterest = [i['options.screener.rule.openInterest'] for i in data]
        self.otm = [i['options.screener.rule.otm'] for i in data]
        self.tobep = [i['options.screener.rule.tobep'] for i in data]
        self.changeRatio = [i['options.screener.rule.changeRatio'] for i in data]
        self.volume = [i['options.screener.rule.volume'] for i in data]
        self.itm = [i['options.screener.rule.itm'] for i in data]
        self.implVol = [i['options.screener.rule.implVol'] for i in data]
        self.close = [i['options.screener.rule.close'] for i in data]
        self.bid = [i['options.screener.rule.bid'] for i in data]

        self.data_dict = { 
            'options.screener.rule.change': self.change,
            'options.screener.rule.expireDate': self.expireDate,
            'options.screener.rule.ask': self.ask,
            'options.screener.rule.openInterest': self.openInterest,
            'options.screener.rule.otm': self.otm,
            'options.screener.rule.tobep': self.tobep,
            'options.screener.rule.changeRatio': self.changeRatio,
            'options.screener.rule.volume': self.volume,
            'options.screener.rule.itm': self.itm,
            'options.screener.rule.implVol': self.implVol,
            'options.screener.rule.close': self.close,
            'options.screener.rule.bid': self.bid
        }

        self.as_dataframe = pd.DataFrame(self.data_dict)


from dataclasses import dataclass, fields
from typing import List, Optional

from discord_webhook import DiscordEmbed


@dataclass
class Day:
    change: float
    change_percent: float
    close: float
    high: float
    last_updated: int
    low: float
    open: float
    previous_close: float
    volume: int
    vwap: float

    @classmethod
    def from_dict(cls, data):
        return cls(**data)


@dataclass
class Details:
    contract_type: str
    exercise_style: str
    expiration_date: str
    shares_per_contract: int
    strike_price: float
    ticker: str

    @classmethod
    def from_dict(cls, data):
        return cls(**data)


@dataclass
class Greeks:
    delta: float
    gamma: float
    theta: float
    vega: float

    @classmethod
    def from_dict(cls, data):
        return cls(**data)


@dataclass
class LastQuote:
    ask: float
    ask_size: int
    bid: float
    bid_size: int
    last_updated: int
    midpoint: float
    timeframe: Optional[str] = None  # adding this new field

    @classmethod
    def from_dict(cls, data: dict) -> 'LastQuote':
        return cls(
            ask=data.get('ask'),
            ask_size=data.get('ask_size'),
            bid=data.get('bid'),
            bid_size=data.get('bid_size'),
            last_updated=data.get('last_updated'),
            midpoint=data.get('midpoint'),
            timeframe=data.get('timeframe')  # account for the new field here too
        )


@dataclass
class LastTrade:
    conditions: List[int]
    exchange: int
    price: float
    sip_timestamp: int
    size: int
    timeframe: str

    @classmethod
    def from_dict(cls, data):
        return cls(**data)


@dataclass
class UnderlyingAsset:
    ticker: Optional[str] = None
    value: Optional[str] = None
    change_to_break_even: Optional[str] = None
    price: Optional[str] = None

    @classmethod
    def from_dict(cls, data):
        # Extract only the keys that are defined in the dataclass
        keys = {f.name for f in fields(cls)}
        data = {k: v for k, v in data.items() if k in keys}
        return cls(**data)



@dataclass
class OptionsSnapshotResult:
    break_even_price: Optional[float] = None
    day: Optional[Day] = None
    details: Optional[Details] = None
    greeks: Optional[Greeks] = None
    ticker: Optional[str] = None
    implied_volatility: Optional[float] = None
    last_quote: Optional[LastQuote] = None
    last_trade: Optional[LastTrade] = None
    open_interest: Optional[int] = None
    underlying_asset: Optional[UnderlyingAsset] = None

    @classmethod
    def from_dict(cls, data: dict) -> 'OptionsSnapshotResult':
        # Extract only the keys that are defined in the dataclass
        keys = {f.name for f in fields(cls)}
        data = {k: v for k, v in data.items() if k in keys}
        
        return cls(
            break_even_price=data.get('break_even_price'),
            day=Day.from_dict(data.get('day')) if data.get('day') else None,
            details=Details.from_dict(data.get('details')) if data.get('details') else None,
            greeks=Greeks.from_dict(data.get('greeks')) if data.get('greeks') else None,
            implied_volatility=data.get('implied_volatility'),
            last_quote=LastQuote.from_dict(data.get('last_quote')) if data.get('last_quote') else None,
            last_trade=LastTrade.from_dict(data.get('last_trade')) if data.get('last_trade') else None,
            open_interest=data.get('open_interest'),
            underlying_asset=UnderlyingAsset.from_dict(data.get('underlying_asset')) if data.get('underlying_asset') else None
        )

    @classmethod
    def from_list(cls, data: List[dict]) -> List['OptionsSnapshotResult']:
        return [cls.from_dict(item) for item in data]
    

    def to_embed(self) -> DiscordEmbed:
        """Create a Discord embed from this article."""
        embed = DiscordEmbed(title=f"Ticker 📰 News", description=f"```py\nOption snapshot result - {self.details.ticker}```")
        # Adding author as a field
        if self.day is not None:
            if self.day.change and self.day.change_percent and self.day.close and self.day.high and self.day.low and self.day.open and self.day.vwap and self.day.previous_close:
                embed.add_embed_field(name="Day Stats:", value=f"> Open: **${self.day.open}**\n> High: **${self.day.high}**\n> Last: **${self.day.close}**\n> Low: **${self.day.low}\n> VWAP: **${self.day.vwap}**\n> Prev. Close: **${self.day.previous_close}**", inline=False)
        
        # Adding publisher name as a field
        if self.details is not None:
            if self.details.contract_type and self.details.exercise_style and self.details.expiration_date and self.details.strike_price and self.details.ticker:
                embed.add_embed_field(name="Option Details:", value=f"> Type: **{self.details.contract_type}**\n> Strike: **${self.details.strike_price}**\n> Expiry: **{self.details.expiration_date}**\n> Style: **{self.details.exercise_style}**\n> Sym: **{self.details.ticker}**")

        
        # Adding published date as a field
        if self.greeks is not None:
            if self.greeks.vega and self.greeks.gamma and self.greeks.delta and self.greeks.theta:
                embed.add_embed_field(name="Greeks:", value=f"> Delta: **{self.greeks.delta}**\n> Gamma: **{self.greeks.gamma}**\n> Vega: **{self.greeks.vega}**\n> Theta: **{self.greeks.theta}**")
        
        if self.last_quote is not None:
            if self.last_quote.ask and self.last_quote.bid and self.last_quote.bid_size and self.last_quote.ask_size and self.last_quote.midpoint:
                embed.add_embed_field(name=f"Last Quote:", value=f"> Bid: **${self.last_quote.bid}**\n> Bid Size: **{self.last_quote.bid_size}**\n> Mid: **${self.last_quote.midpoint}**\n> Ask: **${self.last_quote.ask}**\n> Ask Size: **{self.last_quote.ask_size}**")
        
        # Adding image to embed
        if self.last_trade is not None:
            if self.last_trade.conditions and self.last_trade.exchange and self.last_trade.price and self.last_trade.sip_timestamp and self.last_trade.size:

                embed.add_embed_field(name=f"Last Trade:", value=f"> Size: **{self.last_trade.size}**\n> Price: **${self.last_trade.price}**\n> Exchange: **{self.last_trade.exchange}**\n> Conditions: **{self.last_trade.conditions}**\n> Timestamp: **{self.last_trade.sip_timestamp}**")
        
        if self.underlying_asset is not None:
            if self.underlying_asset.change_to_break_even and self.underlying_asset.ticker:
                value_or_price = self.underlying_asset.value if self.underlying_asset.value else self.underlying_asset.price
                embed.add_embed_field(
                    name="Underlying:",
                    value=f"> Value/Price: **${value_or_price}**\n"
                        f"> Change to Break Even: **{self.underlying_asset.change_to_break_even}**\n"
                        f"> Ticker: **{self.underlying_asset.ticker}**"
                )

        return embed
    


import pandas as pd

class ShortInterest:
    def __init__(self, data):
        self.settlement = [i['settlementDate'] for i in data]
        self.short_int = [i['shortInterst'] for i in data]
        self.avg_volume = [i['avgDailyShareVolume'] for i in data]
        self.days_to_cover = [i['daysToCover'] for i in data]


        self.data_dict = { 

            'settlement_date': self.settlement,
            'short_interest': self.short_int,
            'average_volume': self.avg_volume,
            'days_to_cover': self.days_to_cover
        }



        self.df = pd.DataFrame(self.data_dict)


import os
from dotenv import load_dotenv

load_dotenv()



import asyncio


from apis.polygonio.polygon_options import PolygonOptions


opts = PolygonOptions()




async def main():
    x = await opts.get_near_the_money_data('SPY', exp_greater_than='2023-11-10', exp_less_than='2023-12-15')
    
    # Assuming 'x' is a pandas DataFrame and contains 'expiration' and 'velocity' columns.
    # Group by 'expiration' and then sort each group by 'velocity'
    grouped = x.groupby('expiry', group_keys=False).apply(lambda g: g.sort_values('velocity', ascending=True))
    
    print(grouped)



asyncio.run(main())


import sys
from pathlib import Path
import pytz
# Add the project directory to the sys.path
project_dir = str(Path(__file__).resolve().parents[1])
if project_dir not in sys.path:
    sys.path.append(project_dir)

from typing import List, Optional
from dataclasses import dataclass, field
from polygon_helpers import convert_datetime_list
from polygonio.mapping import stock_condition_dict,STOCK_EXCHANGES
from typing import Dict
from datetime import datetime

class StockSnapshot:
    def __init__(self, data):
        self.ticker = [i.get('ticker') for i in data]
        self.todaysChangePerc = [float(i.get('todaysChangePerc')) for i in data]
        self.todaysChange = [float(i.get('todaysChange')) for i in data]
        self.updated = [i.get('updated') for i in data]



        day = [i.get('day') for i in data]

        # Your existing code with handling for None values
        self.day_o = [float(i.get('o', 0)) for i in day]
        self.day_h = [float(i.get('h', 0)) for i in day]
        self.day_l = [float(i.get('l', 0)) for i in day]
        self.day_c = [float(i.get('c', 0)) for i in day]
        self.day_v = [float(i.get('v', 0)) for i in day]
        self.day_vw = [float(i.get('vw', 0)) for i in day]

        lastQuote = [i.get('lastQuote') or {} for i in data]
        self.ask = [float(i.get('P', 0)) for i in lastQuote]
        self.ask_size = [float(i.get('S', 0)) for i in lastQuote]
        self.bid = [float(i.get('p', 0)) for i in lastQuote]
        self.bid_size = [float(i.get('s', 0)) for i in lastQuote]
        self.quote_timestamp = [self.convert_timestamp(i.get('t')) for i in lastQuote]


        lastTrade = [i.get('lastTrade') for i in data]
        self.trade_conditions = [i.get('c', 0) for i in lastTrade]
        flattened_conditions = [item for sublist in self.trade_conditions for item in (sublist if isinstance(sublist, list) else [sublist])] if self.trade_conditions else []
        self.trade_conditions = [stock_condition_dict.get(c) for c in flattened_conditions] if flattened_conditions is not None else []
        self.trade_id = [i.get('i') for i in lastTrade]
        self.trade_price = [float(i.get('p', 0)) for i in lastTrade]
        self.trade_size = [float(i.get('s', 0)) for i in lastTrade]
        self.trade_timestamp = [self.convert_timestamp(i.get('t')) for i in lastTrade]
        self.trade_exchange = [i.get('x') for i in lastTrade]
        


        min = [i.get('min') for i in data]
        self.min_av = [float(i.get('av')) for i in min]
        self.min_timestamp = [self.convert_timestamp(i.get('t')) for i in min]
        self.min_trades = [float(i.get('n', 0)) for i in min]
        self.min_o = [float(i.get('o', 0)) for i in min]
        self.min_h = [float(i.get('h', 0)) for i in min]
        self.min_l = [float(i.get('l', 0)) for i in min]
        self.min_c = [float(i.get('c', 0)) for i in min]
        self.min_v = [float(i.get('v', 0)) for i in min]
        self.min_vw = [float(i.get('vw', 0)) for i in min]



        prevDay = [i.get('prevDay') for i in data]
        self.o = [float(i.get('o', 0)) for i in prevDay]
        self.h = [float(i.get('h', 0)) for i in prevDay]
        self.l = [float(i.get('l', 0)) for i in prevDay]
        self.c = [float(i.get('c', 0)) for i in prevDay]
        self.v = [float(i.get('v', 0)) for i in prevDay]
        self.vw = [float(i.get('vw', 0)) for i in prevDay]




    def convert_timestamp(self, timestamp):
        if timestamp is None:
            return None
        # Convert nanoseconds to seconds
        timestamp_in_seconds = timestamp / 1_000_000_000
        # Convert to datetime and then to desired string format
        dt = datetime.fromtimestamp(timestamp_in_seconds, pytz.timezone('America/Chicago'))
        return dt.strftime('%Y-%m-%d %H:%M:%S')



import os
from dotenv import load_dotenv
load_dotenv()
from disnake.ext import commands
import disnake
from apis.polygonio.async_polygon_sdk import Polygon
from scripts.rsi_2_4_8 import get_2_4_8_rsi, scan_2_4_8
from disnake import TextInputStyle
import pandas as pd


# Subclassing the modal.
class TaModal(disnake.ui.Modal):
    def __init__(self):
        # The details of the modal, and its components
        components = [
            disnake.ui.TextInput(
                label="Select Your Timespan",
                placeholder="minute,hour,day,week,month,quarter,year",
                custom_id="ta-timespan",
                style=TextInputStyle.short,
                max_length=10,
            ),
            disnake.ui.TextInput(
                label="Select Your Window",
                placeholder="e.g. 14",
                custom_id="ta-window",
                style=TextInputStyle.short,
            ),
            disnake.ui.TextInput(
                label="Select Your Tickers",
                placeholder="e.g. SPY,TSLA,QQQ,MSFT",
                custom_id="ta-tickers",
                style=TextInputStyle.short,
            ),
        ]
        super().__init__(title="Create Tag", components=components)

    # The callback received when the user input is completed.
    async def callback(self, inter: disnake.ModalInteraction):
        user_input = inter.text_values
        tamodal_ = user_input.get('tamodal')


class TACog(commands.Cog):
    def __init__(self, bot):
        self.bot=bot

        self.polygon = Polygon(os.environ.get('TECHNICALS_STRING'))


    @commands.slash_command()
    async def ta(self, inter):
        pass


    @ta.sub_command()
    async def get_all_rsi(self, inter:disnake.AppCmdInter):
        await inter.response.send_modal(TaModal())






    @ta.sub_command()
    async def rsi_2_4_8(self, inter:disnake.AppCmdInter):
        """Scans for daily 2/4/8 that meets criteria"""
        await inter.response.defer()

        results = await scan_2_4_8()



        df = pd.DataFrame(results)
        # Rename the column
        df.columns = ["RESULTS:"]
  
        embed = disnake.Embed(title=f"RSI 2/4/8", description=f"> *This scanner is checking the RSI indicator on the DAILY timeframe for the window-lengths of:    \
        2/4/8*\n\n> **{df}**", color=disnake.Colour.dark_blue())
        embed.add_field(name=f"RSI Conditions:", value=f"> Window 2: **<= 1 & >= 99** ✅\n> Window 4: **<= 10 & >= 90** ✅\n> Window 8: **<= 20 & >= 80 ✅**", inline=False)
   
        await inter.edit_original_message(embed=embed)





def setup(bot:commands.Bot):
    bot.add_cog(TACog(bot))
    print(f"Technicals.. READY!!")


import disnake
import pandas as pd

from disnake.ext import commands
from apis.earnings_whisper.ew_sdk import EarningsWhisper
from apis.polygonio.async_polygon_sdk import Polygon

from discord_.bot_menus.pagination import AlertMenus,PageSelect




class EarningsCOG(commands.Cog):
    def __init__(self, bot):
        self.bot=bot

        self.ew = EarningsWhisper()
        self.poly = Polygon()


    @commands.slash_command()
    async def earnings(self, inter):
        pass


    



    @earnings.sub_command()
    async def sentiment(self, inter:disnake.AppCmdInter):
        """Gets upcoming earnings based on sentiment"""
        await inter.response.defer()
        df = self.ew.get_top_sentiment()

        df = df.as_dataframe
        
        # Convert 'next_er' to datetime if it's not already
        df['next_er'] = pd.to_datetime(df['next_er'])

        # Sort by 'release_time' and 'avg_sentiment'
        df_sorted = df.sort_values(by=['next_er', 'avg_sentiment'], ascending=[True, True])

        embeds = []
        
        for i,row in df_sorted.head(25).iterrows():
            logo = await self.poly.get_polygon_logo(row['ticker'])
            embed = disnake.Embed(title=f"{row['ticker']} || {row['sentiment']}", description=f"```py\n{row['company']} reports earnings at {row['next_er']}```")
            embed.add_field(name=f"Sentiment:", value=f"> **{row['avg_sentiment']}")
            embed.set_thumbnail(logo)
            embed.set_footer(text=f"Sentiment for {row['ticker']}: {row['sentiment']} | Data by EarningsWhisper")
            embeds.append(embed)


        df.to_csv('earnings_sentiment.csv', index=False)

        button = disnake.ui.Button(style=disnake.ButtonStyle.blurple, label='Download', custom_id='buttonSentiment')
        button.callback = lambda interaction: interaction.response.send_message(file=disnake.File('earnings_sesntiment.csv'))

        await inter.edit_original_message(embed=embeds[0], view=AlertMenus(embeds).add_item(EarningsSelect()).add_item(PageSelect(embeds)).add_item(button))

    @earnings.sub_command()
    async def calendar(self, inter:disnake. AppCmdInter, date:str='20231215'):
        """
        Look-up earnings up to a certain date must be in YYYYMMDD!
        """
        await inter.response.defer()
        data = self.ew.calendar(date=date)

        embeds = []
        for i, row in data.as_dataframe.iterrows():
            time_dict = {  
                3: 'afterMarket',
                1: 'preMarket',
                2: 'day'
            }
            company = row['company']
            confirmation_date = row['confirmation_date']
            earnings_time = row['earnings_time']
            next_er_date = row['next_er_date']
            q1_estimate_eps = row['q1_estimate_eps']
            q1_revenue_est = row['q1_revenue_est']
            q_date = row['q_date']
            q_sales = row['q_sales']
            quarter_date = row['quarter_date']
            release_time = row['release_time']
            ticker = row['ticker']
            total = row['total']
            logo = None
            logo = await self.poly.get_polygon_logo(row['ticker'])
            embed = disnake.Embed(title=f"{company} | {next_er_date}", description=f"```py\n{ticker} is expected to report earnings on {next_er_date} - set to report @ {time_dict.get(release_time)}```", color=disnake.Colour.dark_gold())
            embed.add_field(name=f"Revenue EST:", value=f"> **{q1_revenue_est}**")
            embed.add_field(name=f"EPS EST:", value=f"> **{q1_estimate_eps}**")
            embed.set_footer(text=f"{ticker} | ER: {next_er_date} | Data by EarningsWhisper")
            if logo is not None:
                embed.set_thumbnail(logo)
            embeds.append(embed)
        data.as_dataframe.to_csv('earnings.csv', index=False)

        button = disnake.ui.Button(label='Download', style=disnake.ButtonStyle.blurple)
        button.callback = lambda interaction: interaction.response.send_message(file=disnake.File('earnings.csv'))

        await inter.edit_original_message(embed=embeds[0], view=AlertMenus(embeds).add_item(PageSelect(embeds)).add_item(EarningsSelect()).add_item(button))



    @earnings.sub_command()
    async def pivots(self, inter:disnake.AppCmdInter):
        """
        Gets the earnings pivot list with support/resistance.
        """
        await inter.response.defer()
        data = self.ew.pivot_list()

        embeds = []
        for i,row in data.as_dataframe.head(24).iterrows():

            ticker = row['ticker']
            company = row['company']
            list_type = row['list_type']
            last_trade = row['last_trade']
            pivot_point = row['pivot_point']
            resistance_1 = row['resistance_1']
            support_1 = row['support_1']
            resistance_2 = row['resistance_2']
            support_2 = row['support_2']
            resistance_3 = row['resistance_3']
            support_3 = row['support_3']
            logo = await self.poly.get_polygon_logo(ticker)
            embed = disnake.Embed(title=f"Pivot List - {ticker}", description=f"```py\nShowing {company}'s pivot points heading into earnings. This ticker's list type is: {list_type}.```", color=disnake.Colour.dark_gold())

            embed.add_field(name='Support/Res 1', value=f"> R: **{resistance_1}**\n> S: **{support_1}**")

            embed.add_field(name=f'Support/Res 2:', value=f"> R: **{resistance_2}**\n> S: **{support_2}**")
            embed.add_field(name=f"Support/Res 3:", value=f"> R: **{resistance_3}**\n> S: **{support_3}**")
            embed.set_footer(text=f"Pivots: {ticker} | Data by EarningsWhisper")
            embed.set_thumbnail(logo)
            embeds.append(embed)

        data.as_dataframe.to_csv('pivots.csv', index=False)

        await inter.edit_original_message(embed=embeds[0], view=AlertMenus(embeds).add_item(EarningsSelect()).add_item(PageSelect(embeds)), file=disnake.File('pivots.csv'))



    @earnings.sub_command()
    async def today(self, inter:disnake.AppCmdInter):
        """
        Get earnings results for today - revenue, growth, etc.
        
        """

        await inter.response.defer()

        data = self.ew.todays_results()

        df = data.as_dataframe
        embeds = []

        for i, row in df.iterrows():
            earnings_growth = row['earnings_growth']
            earnings_surprise = row['earnings_surprise']
            eps = row['eps']
            er_date = row['er_date']
            estimate = row['estimate']
            ew_grade = row['ew_grade']
            file_name = row['file_name']
            high_estimate = row['high_estimate']
            low_estimate = row['low_estimate']
            name = row['name']
            prev_earnings_growth = row['prev_earnings_growth']
            prev_revenue_growth = row['prev_revenue_growth']
            pwrRating = row['pwrRating']
            quarter = row['quarter']
            revenue = row['revenue']
            revenue_estimate = row['revenue_estimate']
            revenue_growth = row['revenue_growth']
            revenue_surprise = row['revenue_surprise']
            subject = row['subject']
            summary = row['summary']
            ticker = row['ticker']
            whisper = row['whisper']
            logo = await self.poly.get_polygon_logo(ticker)
            embed = disnake.Embed(title=f"{ticker} | Reported: {er_date}", description=f"```py\n{name} Summary for quarter {quarter}:\n{summary}```", color=disnake.Colour.dark_gold())
            embed.add_field(name=f"Estimates:", value=f"> EST: **{estimate}**\n> Low EST: **{low_estimate}**\n> High EST: **{high_estimate}**\n Rev. EST: **{revenue_estimate}**")
            embed.add_field(name=f"Earnings Actual:", value=f"> EPA: **{eps}**\n> Earnings Growth: **{earnings_growth}**\n> Surprise: **{earnings_surprise}**")
            embed.add_field(name=f"Revenue Actual:", value=f"> Revenue: **{revenue}**\n> Growth: **{revenue_growth}**\n> Surprise: **{revenue_surprise}**")
            embed.set_thumbnail(logo)
            embed.add_field(name=f"Previous Growth:", value=f"> Earnings: **{prev_earnings_growth}**\n> Revenue: **{prev_revenue_growth}**")
            embed.add_field(name=f"Grades:", value=f"> PWR: **{pwrRating}**\n> Whisper: **{whisper}**\n> EW Grade: **{ew_grade}**")
            embed.add_field(name=f"Subject:", value=f"> **{subject}**\n> **{file_name}**")
            embed.set_footer(text=f'{ticker} | EPS: {eps} | Data by EarningsWhisper')
            
          
            embed.add_field(name=f"Surprise:", value=f"Earnings: **{earnings_surprise}**\n> Revenue: **{revenue_surprise}**")
            embeds.append(embed)

        file = 'todays_results.csv'
        df.to_csv('todays_results.csv', index=False)


    
        await inter.edit_original_message(embed=embeds[0], view=AlertMenus(embeds).add_item(EarningsSelect()).add_item(PageSelect(embeds)), file=disnake.File(file))



    @earnings.sub_command()
    async def upcoming_etfs(self, inter:disnake.AppCmdInter):
        """
        View upcoming ETFs that have tickers reporting earnings.
        
        """
        await inter.response.defer()
        data = self.ew.upcoming_sectors()

        embeds = []
        for i,row in data.as_dataframe.head(25).iterrows():
            earnings_date = row['earnings_date']
            eps_growth = row['eps_growth']
            max_date = row['max_date']
            revenue_growth = row['revenue_growth']
            sector_id = row['sector_id']
            sector = row['sector']
            sector_name = row['sector_name']
            surprise = row['surprise']
            total = row['total']
            week = row['week']

            embed = disnake.Embed(title=f"Upcoming ETF: | {sector}", description=f"```py\nViewing upcoming earnings for the {sector} sector.", color=disnake.Colour.dark_gold())
            embed.add_field(name=f"Date:", value=f"> **{earnings_date}**")
            embed.add_field(name=f"EPS Growth:", value=f"> **{eps_growth}**")
            embed.add_field(name=f"Surprise:", value=f"> **{surprise}**")
            embed.add_field(name=f"Report Week:", value=f"> **{week}**")
            embed.add_field(name=f"Max Date:", value=f"> **{max_date}**")
            embed.add_field(name=f"Revenue Growth:", value=f"> **{revenue_growth}**")
            embed.add_field(name=f"Total Reporting:", value=f"> **{total}**")

            embeds.append(embed)
            embed.set_footer(text=f'ETF Earnings: {sector} | Data by EarningsWhisper')

            embeds.append(embed)
        data.as_dataframe.to_csv('upcoming_etf_earnings.csv', index=False)
        await inter.edit_original_message(embed=embeds[0], view=AlertMenus(embeds).add_item(EarningsSelect()).add_item(PageSelect(embeds)), file=disnake.File('upcoming_etf_earnings.csv'))
cmds = EarningsCOG(commands.Bot)


class EarningsSelect(disnake.ui.Select):
    def __init__(self):
        super().__init__( 
            placeholder='Choose A Command -->',
            min_values=1,
            max_values=1,
            custom_id='earningCOMMANDselect',
            options = [ 
                disnake.SelectOption(label='Todays Results', value=f'today', description=f'View ranings results on the day.'),
                disnake.SelectOption(label=f"Calendar Date", value='cal', description=f"View earnings up to a certain date YYYYMMDD."),
                disnake.SelectOption(label='Sentiment', value='sentiment', description=f"View upcoming earnings by sentiment rank."),
                disnake.SelectOption(label='Pivot Points', value=f"pivots", description=f"Get support/resistance levels."),
                disnake.SelectOption(label='Upcoming ETFs', value=f"etfs", description=f"ETFs with tickers reporting earnings."),
                disnake.SelectOption(label='Upcoming Russell', value=f"russell", description=f"Upcoming RUSSELL earnings."),

                
            ],
            row=2
        )


    async def callback(self, inter: disnake.AppCommandInter):
        if self.values[0] == 'sentiment':
            await cmds.sentiment(inter)

        elif self.values[0] == 'cal':
            await cmds.calendar(inter, date='20231215')

        elif self.values[0] == 'pivots':
            await cmds.pivots(inter)


        elif self.values[0] == 'today':
            await cmds.today(inter)

        elif self.values[0] == 'etfs':
            await cmds.upcoming_etfs(inter)


def setup(bot: commands.Bot):
    bot.add_cog(EarningsCOG(bot))

    print(f'Earnings commands - ready.')


import pandas as pd

class ETFHoldings:
    def __init__(self, r):
        datalists = r['dataList']
        all_data_dicts = []
        for datalist in datalists:

            self.name = datalist.get('name',None)
            self.changeRatio = datalist.get('changeRatio', None)
            self.shareNumber = datalist.get('shareNumber', None)
            self.ratio = datalist.get('ratio', None)


            self.tickerTuple = datalist.get('tickerTuple', None)
            self.tickerId = datalist.get('tickerId', None)
            self.etfname = datalist.get('name', None)
            self.symbol = datalist.get('symbol', None)



            self.data_dict = { 
                'name': self.name,
                'change_ratio': self.changeRatio,
                'share_number': self.shareNumber,
                'ratio': self.ratio,
                'ticker_id': self.tickerId,
                'etf_name': self.etfname,
                'symbol': self.symbol
            }
            all_data_dicts.append(self.data_dict)

        self.df = pd.DataFrame(all_data_dicts)



from dotenv import load_dotenv
load_dotenv()
import sys
from pathlib import Path
# Add the project directory to the sys.path
project_dir = str(Path(__file__).resolve().parents[1])
if project_dir not in sys.path:
    sys.path.append(project_dir)
from webull_options.webull_options import WebullOptions

db = WebullOptions(database='markets', user='chuck')



import asyncio


async def main():


    await db.db_manager.get_connection()
    all_options = await db.all_options('SPY')
    await db.db_manager.create_table(all_options[0].as_dataframe,table_name='wb_opts', unique_column='option_symbol')
    await db.update_all_options()


asyncio.run(main())


import os
from dotenv import load_dotenv
import aiohttp
import pandas as pd
import asyncio
import matplotlib.pyplot as plt
load_dotenv()
from fudstop.apis.webull.webull_options import WebullOptions
from fudstop.apis.helpers import get_human_readable_string

wb = WebullOptions(os.getenv('WEBULL_OPTIONS'))

async def associate_dates_with_data(dates, datas):
    if datas is not None and dates is not None:
    # This function remains for your specific data handling if needed
        return [{**data, 'date': date} for date, data in zip(dates, datas)]

async def fetch_volume_analysis(session, option_symbol, id, underlying_ticker):
    url = f"https://quotes-gw.webullfintech.com/api/statistic/option/queryVolumeAnalysis?count=200&tickerId={id}"
    async with session.get(url) as resp:
        if resp.status == 200:
            vol_anal = await resp.json()
            dates = vol_anal.get('dates')
            datas = vol_anal.get('datas')
            associated_data = await associate_dates_with_data(dates, datas)

            df = pd.DataFrame(associated_data)
            df['option_symbol'] = option_symbol
            components = get_human_readable_string(option_symbol)
            df['underlying_ticker'] = underlying_ticker
            df['strike'] = components.get('strike_price')
            df['call_put'] = components.get('call_put')
            df['expiry'] = components.get('expiry_date')

            return df
        else:
            print(f"Failed to fetch data for ID {id}: HTTP Status {resp.status}")
            return pd.DataFrame()

async def main(ticker, date:str='2023-12-15'):
    headers = wb.headers
    option_ids = await wb.get_option_ids(ticker)  # Adjust based on actual return structure

    async with aiohttp.ClientSession(headers=headers) as session:
        tasks = [fetch_volume_analysis(session, option_symbol, id, underlying_ticker) for option_symbol, id, underlying_ticker in option_ids]
        dfs = await asyncio.gather(*tasks)  # Run all tasks concurrently
        
    final_df = pd.concat(dfs, ignore_index=True)
    print(final_df)
    final_df.to_csv(f'volume_analysis.csv', index=False)

    # Define the visualization function
    def plot_volume_by_strike_for_date(df, target_date, dark_background=True):
        # Extract data for the specific date
        target_df = df[df['expiry'] == pd.to_datetime(target_date)]
        
        # If no data available for the target_date, inform and return without plotting
        if target_df.empty:
            print(f"No data available for the date: {target_date}")
            return

        plt.style.use('dark_background' if dark_background else 'default')
        
        # Create a plot with dark background
        fig, ax = plt.subplots(figsize=(12, 6), facecolor='black' if dark_background else 'white')
        
        # Plot buy and sell volumes
        ax.bar(target_df['strike'], target_df['buy'], label='Buy Volume', color='green', alpha=0.7)
        ax.bar(target_df['strike'], -target_df['sell'], label='Sell Volume', color='red', alpha=0.7)
        
        # Set title and axes labels with gold color and high readability
        ax.set_title(f'Buy/Sell Volume by Strike for Expiry: {pd.to_datetime(target_date).strftime("%Y-%m-%d")}', color='gold')
        ax.set_xlabel('Strike Price', color='gold')
        ax.set_ylabel('Volume', color='gold')
    
        # Tick parameters for readability
        ax.tick_params(axis='x', colors='gold')
        ax.tick_params(axis='y', colors='gold')

        # Adding grid for better readability, with a lighter color
        ax.grid(True, color='dimgray')
        
        # Show legend with readable font color
        ax.legend(facecolor='darkgray', edgecolor='gold', fontsize='large')
        
        # Show the plot
        plt.show()

    # Example usage:
    # Load the data (make sure to parse the 'date' and 'expiry' columns correctly)
    df = pd.read_csv('volume_analysis.csv', parse_dates=['date', 'expiry'])

    # Filter out rows with missing values in the volume, strike, or expiry columns if necessary
    # df = df.dropna(subset=['buy_volume', 'sell_volume', 'strike', 'expiry'])

    # Call the function with the chosen date
    plot_volume_by_strike_for_date(df, date, dark_background=True)


import os
from fudstop.scripts.rsi_screener import gather_rsi_results
min1_td9 = os.environ.get("min1_td9")
min5_td9 = os.environ.get("min5_td9")
min15_td9 = os.environ.get("min15_td9")
min20_td9 = os.environ.get("min20_td9")
min30_td9 = os.environ.get("min30_td9")
min60_td9 = os.environ.get("min60_td9")
min120_td9 = os.environ.get("min120_td9")
min240_td9 = os.environ.get("min240_td9")
day_td9 = os.environ.get("day_td9")
min2_td9 = os.environ.get("min2_td9")
min3_td9 = os.environ.get("min3_td9")
min10_td9 = os.environ.get("min10_td9")
week_td9 = os.environ.get("week_td9")
access_token = os.environ.get("ACCESS_TOKEN")
month_td9 = os.environ.get("month_td9")

import asyncio

asyncio.run(gather_rsi_results(timespan='minute', webhook_url=min10_td9))


import sys
from pathlib import Path

# Add the project directory to the sys.path
project_dir = str(Path(__file__).resolve().parents[1])
if project_dir not in sys.path:
    sys.path.append(project_dir)

from polygon.websocket import WebSocketMessage
from apis.helpers import convert_to_ns_datetime
import asyncio
from typing import List



"""
>>> Handles messages from the polygon.io websocket cluster 

"""

async def indices_handler(msg: WebSocketMessage, data_queue: asyncio.Queue, db_manager=None):
    for m in msg:
    # Fetch the name using the incoming symbol

            # database_data = { 
            #     'name': name,
            #     'ticker': m.symbol,
            #     'day_open': m.official_open_price,
            #     'minute_open': m.open,
            #     'minute_high': m.high,
            #     'minute_low': m.low,
            #     'minute_close': m.close,
            #     'minute_start': convert_to_ns_datetime(m.start_timestamp),
            #     'minute_end': convert_to_ns_datetime(m.end_timestamp)
            # }

            data_queue_data = { 
                'type': 'Indices',
                'ticker': m.symbol,
                'day_open': m.official_open_price,
                'minute_open': m.open,
                'minute_high': m.high,
                'minute_low': m.low,
                'minute_close': m.close,
                'minute_start': convert_to_ns_datetime(m.start_timestamp),
                'minute_end': convert_to_ns_datetime(m.end_timestamp)
            }

            # if db_manager is not None:
            #     asyncio.create_task(db_manager.save_structured_message(data, "indices_aggs"))

            #await db.insert_indices_aggs(data)
            await data_queue.put(data_queue_data)
        


import os
from dotenv import load_dotenv
load_dotenv()
import matplotlib.pyplot as plt
import disnake
import plotly.graph_objects as go
from selenium import webdriver
from selenium.webdriver.chrome.options import Options
import aiohttp
from plots import plot_calls_and_puts,plot_greek_exposure,plot_iv_surface
from disnake.ext import commands
from discord_.bot_menus.pagination import AlertMenus, PageSelect
from discord_.autocomp import ticker_autocomp,strike_autocomp,expiry_autocomp
from apis.polygonio.polygon_options import PolygonOptions
from apis.webull.webull_options import WebullOptions
from apis.webull.webull_trading import WebullTrading
from apis.polygonio.polygon_database import PolygonDatabase
from tabulate import tabulate
from all_helpers import determine_emoji
import asyncio
from apis.helpers import format_large_numbers_in_dataframe
import pandas as pd


class OptionsCOG(commands.Cog):
    def __init__(self, bot):
        self.bot=bot
        self.poly = PolygonOptions(user='chuck', database='markets')
        self.wt = WebullTrading()
        self.wb = WebullOptions(os.environ.get('WEBULL_OPTIONS'))
        self.db = PolygonDatabase(user='chuck', database='markets')



    @commands.slash_command()
    async def options(self, inter):
        pass


    @options.sub_command()
    async def full_skew(self, inter:disnake.AppCmdInter, ticker:str):
        """Gets the lowest IV call/put option for all expirations for a ticker"""
        ticker = ticker.upper()
        await inter.response.defer()
        full_skew, price = await self.poly.full_skew(ticker)
        full_skew['iv'] = (full_skew['iv'] * 100).round(2)
        full_skew = full_skew.rename(columns={'strike': 'skew_strike'})
        full_skew = full_skew.drop(columns=['call_put'])
        table = tabulate(full_skew, headers='keys', tablefmt='fancy', showindex=False)


        chunks = [table[i:i + 3860] for i in range(0, len(table), 3860)]
        embeds=[]
        for chunk in chunks:
            embed = disnake.Embed(title=f"Full Skew - {ticker}", description=f"> Current Price: **${price}**\n```py\n{chunk}```", color=disnake.Colour.dark_orange())
            embed.add_field(name=f"Full Skew for: {ticker}",value=f"> Price: **{price}**")
            embed.set_footer(text=f'Full Skew: {ticker} | Data by Polygon.io | Implemented by FUDSTOP')
            embeds.append(embed)

        button = disnake.ui.Button(style=disnake.ButtonStyle.blurple, label='Download')
        button.callback = lambda interaction: interaction.response.send_message(file=disnake.File('full_skew.csv'))

        
        await inter.edit_original_message(embed=embeds[0], view=AlertMenus(embeds).add_item(PageSelect(embeds)).add_item(button).add_item(OptsSelect(ticker)))


    @options.sub_command()
    async def top_vol_strikes(self, inter:disnake.AppCmdInter, ticker:str):
        """
        Gets the top volume strikes across all expirations for calls/puts.
        
        """
        await inter.response.defer()
        ticker = ticker.upper()

        data = await self.poly.vol_oi_top_strike(ticker)

        # Ensure 'expiry' is a datetime column
        data['expiry'] = pd.to_datetime(data['expiry'])

        # Change date format to 'YY-MM-DD'
        data['expiry'] = data['expiry'].dt.strftime('%y-%m-%d')

        data = format_large_numbers_in_dataframe(data)
        data = data.drop(columns=['ticker'])
        data.to_csv('top_vol_strikes.csv', index=False)
        table = tabulate(data, headers='keys', tablefmt='fancy', showindex=False)


        chunks = [table[i:i + 3860] for i in range(0, len(table), 3860)]
        embeds=[]
        for chunk in chunks:
            embed = disnake.Embed(title=f"TopVol Strikes - {ticker}", description=f"```py\n{chunk}```", color=disnake.Colour.dark_orange())
            embed.set_footer(text=f'TopVol Strikes: {ticker} | Data by Polygon.io | Implemented by FUDSTOP')
            embeds.append(embed)

        button = disnake.ui.Button(style=disnake.ButtonStyle.blurple, label='Download')
        button.callback = lambda interaction: interaction.response.send_message(file=disnake.File('top_strike_vol.csv'))

        
        await inter.edit_original_message(embed=embeds[0], view=AlertMenus(embeds).add_item(PageSelect(embeds)).add_item(button).add_item(OptsSelect(ticker)))



    @options.sub_command()
    async def volume_analysis(self, inter:disnake.AppCmdInter, ticker:str, date:str):
        """Enter a date in YYYY-MM-DD and view the volume analysis per strike!"""
        ticker =ticker.upper()
        await inter.response.defer()

        
        headers = self.wb.headers
        option_ids = await self.wb.get_option_ids(ticker)  # Adjust based on actual return structure

        async with aiohttp.ClientSession(headers=headers) as session:
            tasks = [self.wb.fetch_volume_analysis(session, option_symbol, id, underlying_ticker) for option_symbol, id, underlying_ticker in option_ids]
            dfs = await asyncio.gather(*tasks)  # Run all tasks concurrently
            
        final_df = pd.concat(dfs, ignore_index=True)
        print(final_df)
        price = await self.wt.stock_quote(ticker)
        price_ = price.web_stock_close
        final_df['underlying_price'] = price_
        final_df = final_df.sort_values('buy', ascending=False)
        final_df.to_csv(f'volume_analysis.csv', index=False)

        def plot_volume_by_strike_for_date(df, target_date, option_type='all', dark_background=True):
            # Extract data for the specific date
            target_df = df[(df['expiry'] == pd.to_datetime(target_date)) & (df['call_put'] == option_type)]
            
            # If no data available for the target_date, inform and return without plotting
            if target_df.empty:
                print(f"No data available for the date: {target_date} and type: {option_type}")
                return

            plt.style.use('dark_background' if dark_background else 'default')
            
            # Create a plot with dark background
            fig, ax = plt.subplots(figsize=(12, 6), facecolor='black' if dark_background else 'white')
            
            # Plot buy and sell volumes for calls and puts separately
            calls = target_df[target_df['option'] == 'call']
            puts = target_df[target_df['option'] == 'put']
            
            width = (target_df['strike'].max() - target_df['strike'].min()) / len(target_df['strike'].unique()) * 0.4
            
            # Call Volumes
            ax.bar(calls['strike'] - width/2, calls['buy'], width=width, label='Call Buy Volume', color='blue', alpha=0.7)
            ax.bar(calls['strike'] - width/2, -calls['sell'], width=width, label='Call Sell Volume', color='lightblue', alpha=0.7)
            
            # Put Volumes
            ax.bar(puts['strike'] + width/2, puts['buy'], width=width, label='Put Buy Volume', color='purple', alpha=0.7)
            ax.bar(puts['strike'] + width/2, -puts['sell'], width=width, label='Put Sell Volume', color='pink', alpha=0.7)
            
            # Set title and axes labels with gold color and high readability
            ax.set_title(f'Call/Put Buy/Sell Volume by Strike for Expiry: {pd.to_datetime(target_date).strftime("%Y-%m-%d")}', color='gold')
            ax.set_xlabel('Strike Price', color='gold')
            ax.set_ylabel('Volume', color='gold')
            
            # Tick parameters for readability
            ax.tick_params(axis='x', colors='gold')
            ax.tick_params(axis='y', colors='gold')

            # Adding grid for better readability, with a lighter color
            ax.grid(True, color='dimgray')
            
            # Show legend with readable font color
            ax.legend(facecolor='darkgray', edgecolor='gold', fontsize='large')
            
            # Show the plot
            plt.tight_layout()
            plt.savefig('plot.png')
            plt.show()

        # Example usage:
        # Load the data (make sure to parse the 'date' and 'expiry' columns correctly)
        df = pd.read_csv('volume_analysis.csv', parse_dates=['date', 'expiry'])

        # Filter out rows with missing values in the volume, strike, or expiry columns if necessary
        # df = df.dropna(subset=['buy_volume', 'sell_volume', 'strike', 'expiry'])

        # Call the function with the chosen date
        plot_volume_by_strike_for_date(df, date, dark_background=True)
        file = disnake.File('plot.png', filename='plot.png')
        embed = disnake.Embed(title=f"Volume Analysis - {ticker} | {date}")
        
        embed.set_image(url="attachment://plot.png")

        await inter.edit_original_message(embed=embed, file=file)


    @options.sub_command()
    async def all(self, inter:disnake.AppCmdInter, ticker:str):
        
        """Returns a CSV file of all options and accompanying data for a ticker."""
        await self.poly.connect()
        if ticker == 'SPX':
            ticker = 'I:SPX'
        elif ticker == 'NDX':
            ticker = 'I:NDX'
        elif ticker == 'VIX':
            ticker = 'I:VIX'
        await inter.response.defer(ephemeral=False)
        ticker = ticker.upper()

        all_options = await self.poly.get_option_chain_all(ticker)
        
        all_options.df.to_csv('all_options.csv', index=False)
        df = all_options.df
        df.columns = df.columns.str.lower()
        df = df.rename(columns={'ticker': 'symbol', 'oi': 'open_interest', 'cp': "call_put", 'exp': 'expiry', 'vol': 'volume', 'entrycost': "'entry_cost'"})
        df = df.drop(columns=['conditions', 'exchange'])

        await self.poly.batch_insert_dataframe(df, table_name='optionsall', unique_columns='symbol, strike, expiry, call_put')
        embed = disnake.Embed(title=f"All Options - {ticker}", description=f'```py\nYour downloadable CSV is ready!```', color=disnake.Colour.brand_green())
        embed.add_field(name=f"Info:", value=f"This file contains all options for {ticker} across ALL expiration dates and contains ALL accompanying data.")
        embed.set_footer(text="Data Provided by Polygon.IO | Implemented by FUDSTOP", icon_url=os.environ.get('fudstop'))
        await inter.edit_original_message(embed=embed,file=disnake.File('all_options.csv'), view=OptionsView(ticker).add_item(OptsSelect(ticker)))







        return df
    







    # @options.sub_command()
    # async def snapshot(self, inter:disnake.AppCmdInter, ticker:str=commands.Param(autocomplete=ticker_autocomp), strike:str=commands.Param(autocomplete=strike_autocomp), expiry:str=commands.Param(autocomplete=expiry_autocomp), call_put: str=commands.Param(choices=["Call", "Put"])):
    #     """View the snapshot of an option contract!"""
    #     await inter.response.defer()
    #     ticker = build_option_symbol(underlying_symbol=ticker,strike_price=strike,call_or_put=call_put,expiry=expiry,prefix_o=True)
    #     print(ticker)
    #     snapshot_data = await master.get_universal_snapshot(ticker)
    #     if snapshot_data.change_percent and snapshot_data.change_percent[0] is not None:  # Check if the list is not empty
    #         if snapshot_data.change_percent[0] > 0:
    #             color = disnake.Colour.dark_green()
    #         elif snapshot_data.change_percent[0] < 0:
    #             color = disnake.Colour.dark_red()
    #         else:
    #             color = disnake.Colour.dark_grey()
    #             print("change_percent is empty")
    #     if snapshot_data.volume[0] is not None and snapshot_data.open_interest[0] is not None:
    #         if snapshot_data.volume[0] > snapshot_data.open_interest[0]:
    #             # Your logic here
    #             unusual = 'YES'
    #         else:
    #             unusual = 'NO'
    #         # Handle the case where one or both of the values are None
    #         print("Either volume or open_interest or both are None.")
    #     else:
    #         unusual = 'N/A'
            

    #     embed = disnake.Embed(title=f"Option Snapshot - {snapshot_data.underlying_ticker[0]}", description=f"```py\nViewing the most up-to-date option data for {snapshot_data.underlying_ticker[0]} | {snapshot_data.strike[0]} | {snapshot_data.expiry[0]} | {snapshot_data.contract_type}.```\n> *Color of the embed is represented by the change % of the contract on the day.*", color=color)
    #     embed.add_field(name=f"Day:", value=f"> Open: **${snapshot_data.open[0]}\n> High: **${snapshot_data.high[0]}**\n> Low: **${snapshot_data.low[0]}**\n> Close: **${snapshot_data.close[0]}**\n> PrevClose: **${snapshot_data.prev_close[0]}**\n> Change%: **{round(float(snapshot_data.change_percent[0]),2)}%**", inline=False)
    #     try:
    #         embed.add_field(name=f"Greeks:", value=f"> Delta: **{round(float(snapshot_data.delta[0]),2)}**\n> Gamma: **{round(float(snapshot_data.gamma[0]),2)}**\n> Theta: **{round(float(snapshot_data.theta[0]),2)}**\n> Vega: **{round(float(snapshot_data.vega[0]),2)}**")
    #     except TypeError:
    #         embed.add_field(name=f"Greeks:", value=f"> Delta: **{snapshot_data.delta}**\n> Gamma: **{snapshot_data.gamma}**\n> Theta: **{snapshot_data.theta}**\n> Vega: **{snapshot_data.vega}**")
    #     embed.add_field(name=f"Vol. Vs OI:", value=f"> VOL: **{float(snapshot_data.volume[0]):,}**\n> OI: **{float(snapshot_data.open_interest[0]):,}**\n> Unusual? **{unusual}**")
    #     embed.add_field(name=f"Last Trade:", value=f"> Size: **{snapshot_data.trade_size[0]}**\n> Price: **${snapshot_data.trade_price[0]}**\n> Exchange: **${snapshot_data.exchange[0]}**\n> Conditions: **{snapshot_data.conditions[0]}**")
    #     embed.add_field(name=f"Last Quote:", value=f"> Bid: **{snapshot_data.bid[0]}**\n> Bid Size: **{snapshot_data.bid_size[0]}**\n> Mid: **${snapshot_data.midpoint[0]}**\n> Ask: **{snapshot_data.ask[0]}**\n> Ask Size: **{snapshot_data.ask_size[0]}**")
    #     embed.set_footer(text=f"{ticker} | Data Provided by Polygon.io | Implemented by FUDSTOP")
  


    #     await inter.edit_original_message(embed=embed, view=OptionsView(ticker, self.bot,snapshot_data.data_dict))



    @options.sub_command()
    async def allskew(self, inter:disnake.AppCmdInter):
        """Scans and returns all skews with a depth of 5 or more, or -5 or less."""
        await inter.response.defer()


        tasks = [self.db.process_ticker(ticker) for ticker in self.wb.most_active_tickers]
        results = await asyncio.gather(*tasks)

        # Filter out None values and print the first non-None result
        valid_results = [result for result in results if result is not None]
        

        df = pd.DataFrame(valid_results)

        df['expiration'] = pd.to_datetime(df['expiration']).dt.strftime('%y/%m/%d')
        # Apply the function to add a new column to the DataFrame
        df['dir'] = df.apply(determine_emoji, axis=1)

        # Tabulate the DataFrame
        tabulated_data = tabulate(df, headers='keys', tablefmt='fancy', showindex=False)
        


        chunks = [tabulated_data[i:i + 4000] for i in range(0, len(tabulated_data), 4000)]
        embeds=[]
        for chunk in chunks:
            embed = disnake.Embed(title=f"All Skew", description=f"```py\n{chunk}```", color=disnake.Colour.dark_purple())
            embed.add_field(name=f"_ _", value=f"> **Viewing all skews.**")
            embeds.append(embed)

        await inter.edit_original_message(embed=embeds[0], view=AlertMenus(embeds))


    @options.sub_command()
    async def all_trades(self, inter:disnake.AppCmdInter, ticker:str=commands.Param(autocomplete=ticker_autocomp)):
        """View all option trades for an underlying ticker - ordered by volume"""
        await self.poly.connect()
        await inter.response.defer()

        query = f"""SELECT underlying_symbol, strike, call_put, expiry,size, price FROM optiontrade WHERE underlying_symbol = '{ticker}' and expiry > '2023-10-04' ORDER BY size DESC LIMIT 5000"""

        results = await self.poly.fetch(query)



        df = pd.DataFrame(results, columns=['sym', 'strike', 'call/put', 'exp', 'size', 'price'])
        table = tabulate(df, headers='keys', tablefmt='fancy', showindex=False)

        chunks = [table[i:i + 4000] for i in range(0, len(table), 4000)]
        embeds=[]
        for chunk in chunks:
            embed = disnake.Embed(title=f"All Trades: {ticker}", description=f"```py\n{chunk}```", color=disnake.Colour.dark_purple())
            embed.add_field(name=f"Viewing all option trades for:", value=f"> **{ticker}**")
            embeds.append(embed)

        await inter.edit_original_message(embed=embeds[0], view=AlertMenus(embeds))


    @options.sub_command()
    async def iv_skew(self, inter:disnake.AppCmdInter, ticker: str):
        # Step 1: Fetch Data
        ticker = ticker.upper()
        await inter.response.defer()
        all_options = await opts.get_option_chain_all(ticker)
        all_options = all_options.df.sort_values('strike')
        all_options.to_csv('all_options.csv', index=False)
        # Aggregate IV by strike (taking average for this example)
        # Sort by IV and take the first (smallest) value
        min_iv_row = all_options.sort_values('iv', ascending=True).iloc[0]
        full_iv_sorted = all_options.sort_values('iv', ascending=True)

        min_iv_value = min_iv_row['iv']
        min_iv_strike = min_iv_row['strike']


        # Step 2: Visualize Data with Plotly
        # Now plot
        # Create the figure
        fig = go.Figure()

        # Add the IV vs Strike curve
        fig.add_trace(go.Scatter(
            x=full_iv_sorted['strike'],
            y=full_iv_sorted['iv'],
            mode='lines+markers',
            line=dict(color='blue', width=2.0),  # Vibrant blue line with increased width
            marker=dict(size=8, color='purple', opacity=0.7)  # Red markers
        ))

        # Extract the UNDERLYING_PRICE (assuming it's a constant for all rows, hence taking the first row's value)
        underlying_price = all_options.iloc[0]['price']

        # Update the title to include the underlying price
        fig.update_layout(
            title=f"{ticker} Consolidated IV Skew | Underlying Price: ${underlying_price:.2f}",  # Format it to 2 decimal places
            titlefont=dict(size=20),
            xaxis_title="Strike",
            yaxis_title="Implied Volatility",
            template="plotly_dark"
        )

        # If you want to add it as an annotation next to the lowest IV strike:
        if min_iv_strike > underlying_price:
            result = "🟢"
            color="green"
        else:
            result = "🔥"
            color="red"
        fig.add_annotation(
            x=min_iv_strike,
            y=min_iv_value,
            xref="x",
            yref="y",
            text=f"Price: ${underlying_price:.2f} vs. Low IV: ${min_iv_strike} == {result}",
            showarrow=True,
            arrowhead=4,
            ax=0,
            ay=-160,  # Adjust the y-offset to position it correctly
            font=dict(size=18, color=color),
            bordercolor="#c7c7c7",
            borderwidth=2,
            borderpad=1,
            bgcolor="black",
            opacity=0.8
        )
        fig.write_html('temp_plot_options.html')

        # Step 3: Convert HTML to Image with Selenium
        file_path = os.path.abspath('temp_plot_options.html')
        options = webdriver.ChromeOptions()
        options.add_argument('--headless')
        browser = webdriver.Chrome(options=options)
        browser.get(f'file://{file_path}')
        screenshot_bytes = browser.get_screenshot_as_png()
        with open('screenshot2.png', 'wb') as f:
            f.write(screenshot_bytes)
        browser.quit()

        with open('screenshot2.png', "rb") as f:
            file = disnake.File(f, filename="screenshot2.png")  # Use file pointer directly

        embed = disnake.Embed().set_image(url="attachment://screenshot2.png")
        embed.add_field(name=f"Chart:", value=f"> **Viewing consolidated IV skew across all options for {ticker}. This is a visual representation of the consolidated skew across all expirations.")
        embed.set_footer(text=f'IV Skew - {ticker} | Data by Polygon.Io | Implemented by FUDSTOP', icon_url=os.environ.get('fudstop_logo'))
        await inter.edit_original_message(file=file, embed=embed, view = OptionsView(ticker).add_item(OptsSelect(ticker)))


    @options.sub_command()
    async def iv_surface(self, inter:disnake.AppCmdInter, ticker:str):
        """Plots the IV surface of a ticker across all expirations and strikes."""
        ticker = ticker.upper()
        await inter.response.defer()

        await plot_iv_surface(ticker)

        
        with open('iv_surface.png', "rb") as f:
            file = disnake.File(f, filename="iv_surface.png")  # Use file pointer directly

        embed = disnake.Embed().set_image(url="attachment://iv_surface.png")
        embed.add_field(name=f"Chart:", value=f"> **Viewing IV Surface 3D Plot for {ticker}. This is a visual representation of the volatility surface across all strikes and expirations - showing days until expiry vs strike + IV levels.")
        embed.set_footer(text=f'IV Surface - {ticker} | Data by Polygon.Io | Implemented by FUDSTOP', icon_url=os.environ.get('fudstop'))
        await inter.edit_original_message(file=file, embed=embed, view = OptionsView(ticker))




    @options.sub_command()
    async def greek_exposure(self, inter:disnake.AppCmdInter, ticker:str, greek:str=commands.Param(choices=['delta', 'gamma', 'theta', 'vega'])):
        """Charts the delta/gamma/vega/theta exposure across all strikes and expirations."""

        await inter.response.defer()
        ticker = ticker.upper()
        await plot_greek_exposure(ticker, greek)
        
        with open(f'files/{greek}.jpg', "rb") as f:
            file = disnake.File(f, filename=f"files/{greek}.jpg")  # Use file pointer directly

        embed = disnake.Embed().set_image(url=f"attachment://files/{greek}.jpg")
        embed.add_field(name=f"Chart:", value=f"> **Viewing consolidated vega across all strikes for {ticker}. This is a visual representation of the current levels of VEAG per strike on a consolidated basis across all expirations.")
        embed.set_footer(text=f'Consolidated {greek.upper()} - {ticker} | Data by Polygon.Io | Implemented by FUDSTOP', icon_url=os.environ.get('fudstop'))
        await inter.edit_original_message(file=file, embed=embed, view = OptionsView(ticker).add_item(OptsSelect(ticker)))

    


opts = OptionsCOG(commands.Bot)

class GreekModal(disnake.ui.Modal):
    def __init__(self, greek):
        self.greek=  greek
        

        components = [

        disnake.ui.TextInput(
            label=f"Choose a ticker..",
            placeholder='e.g. AAPL',
            custom_id=f"greekmodal",
            style=TextInputStyle.short,
            max_length=6,
            required=True
        ),
        
        ]
            
        # Make sure to pass the components to the super().__init__
        super().__init__(title="Query Options Database", components=components)
        
    async def callback(self, inter: disnake.ModalInteraction):
        await inter.response.defer()
        user_input = inter.text_values
        ticker = user_input.get('greekmodal')

        await plot_greek_exposure(ticker, greek=self.greek)

        with open(f'files/{self.greek}.jpg', "rb") as f:
            file = disnake.File(f, filename=f"files/{self.greek}.jpg")  # Use file pointer directly

        embed = disnake.Embed().set_image(url=f"attachment://files/{self.greek}.jpg")
        embed.add_field(name=f"Chart:", value=f"> **Viewing consolidated vega across all strikes for {ticker}. This is a visual representation of the current levels of VEAG per strike on a consolidated basis across all expirations.")
        embed.set_footer(text=f'Consolidated {self.greek.upper()} - {ticker} | Data by Polygon.Io | Implemented by FUDSTOP', icon_url=os.environ.get('fudstop'))
        await inter.edit_original_message(file=file, embed=embed, view = OptionsView(ticker).add_item(OptsSelect(ticker)))

    









class OptsSelect(disnake.ui.Select):
    def __init__(self, ticker=None):
        self.ticker=ticker

        super().__init__( 
            placeholder='Choose a command -->',
            min_values=1,
            max_values=1,
            custom_id='optsCOGselect',
            options = [ 
                disnake.SelectOption(label='Analyze',value='9', description=f'Analyze the options market..'),
                disnake.SelectOption(label='Full Skew',value='0', description=f'View the skew across all expirations.'),
                disnake.SelectOption(label='Top Vol. Strikes',value='1', description=f'View top volume strikes across all expirations.'),
                disnake.SelectOption(label='Greek Exposure', value='3', description='View the consolidated greek exposure of choice.')
            ]
        )


    async def callback(self, inter:disnake.AppCmdInter):
        if self.values[0] == '0':
            await opts.full_skew(inter, self.ticker)

        elif self.values[0] == '1':
            await opts.top_vol_strikes(inter, self.ticker)
 
        elif self.values[0] == '2':
            await opts.lowest_theta(inter, self.ticker)

        elif self.values[0] == '3':
            await opts.gamma_exposure(inter, self.ticker)

        elif self.values[0] == '9':
            await inter.response.edit_message(view=OptionsView().add_item(OptsSelect(self.ticker)))

from disnake import TextInputStyle

# Subclassing the modal.
class MyModal(disnake.ui.Modal):
    def __init__(self):
        # The details of the modal, and its components
        components = [
            disnake.ui.TextInput(
                label="Ticker",
                placeholder="e.g. AAPL",
                custom_id="ticker",
                style=TextInputStyle.short,
                max_length=10,
            ),
            disnake.ui.TextInput(
                label="Strike",
                placeholder="e.g. 125",
                custom_id="strike",
                style=TextInputStyle.short,
                max_length=10
            ),
            disnake.ui.TextInput(
                label="Call or Put",
                placeholder="e.g. call",
                custom_id="call_put",
                style=TextInputStyle.short,
                max_length=4
            ),
            disnake.ui.TextInput(
                label="Expiration Date",
                placeholder="e.g. 231208",
                custom_id="expiry",
                style=TextInputStyle.short,
                max_length=6
            ),
        ]
        super().__init__(title="Select Your Option", components=components)

    # The callback received when the user input is completed.
    async def callback(self, inter: disnake.ModalInteraction):
   
        embed = disnake.Embed(title="Option Selector", description=f"```py\nThis is a test.```")
        expiry = inter.text_values.get('expiry')
        strike = inter.text_values.get('strike')
        call_put = inter.text_values.get('call_put')
        ticker = inter.text_values.get('ticker')
        await inter.send(embed=embed, ephemeral=True, view=OptionsView(ticker=ticker, strike=strike, call_put=call_put, expiry=expiry))


class OptionsView(disnake.ui.View):
    def __init__(self, greek:str='vega', ticker=None, strike=None, call_put=None, expiry=None):
        self.ticker=ticker
        self.strike=strike
        self.call_put=call_put
        self.greek=greek
        self.expiry=expiry
        super().__init__(timeout=None)
        # Add the button only if all the option arguments are passed in
        if all([self.ticker, self.strike, self.call_put, self.expiry]):
            self.add_item(disnake.ui.Button(style=disnake.ButtonStyle.blurple, label='OPTION LOADED', row=4, custom_id='optview2'))
            self.remove_item(self.optview1)

    @disnake.ui.button(style=disnake.ButtonStyle.blurple, label='Pick an Option', row=0, custom_id='optview1')
    async def optview1(self, button:disnake.ui.Button, inter:disnake.AppCmdInter):
  
        await inter.response.send_modal(MyModal())



    @disnake.ui.button(style=disnake.ButtonStyle.red, row=0, custom_id='theta', emoji='⏳', label='Theta')
    async def theta(self, button: disnake.ui.Button, inter: disnake.AppCmdInter):

        await inter.response.send_modal(GreekModal(greek=self.greek))


    @disnake.ui.button(style=disnake.ButtonStyle.blurple, row=0, custom_id='gamma', emoji='🎯', label='Gamma')
    async def gamma(self, button: disnake.ui.Button, inter: disnake.AppCmdInter):

        await inter.response.send_modal(GreekModal(greek=self.greek))



    @disnake.ui.button(style=disnake.ButtonStyle.grey, row=1, custom_id='delta', emoji='⚔️', label='Delta')
    async def delta(self, button: disnake.ui.Button, inter: disnake.AppCmdInter):

        await inter.response.send_modal(GreekModal(greek=self.greek))



    @disnake.ui.button(style=disnake.ButtonStyle.green, row=1, custom_id='vega', emoji='✨', label='Vega')
    async def vega(self, button: disnake.ui.Button, inter: disnake.AppCmdInter):

        await inter.response.send_modal(GreekModal(greek=self.greek))





        




def setup(bot: commands.Bot):
    bot.add_cog(OptionsCOG(bot))
    print(f'Options command - Ready')


import sys
import os

sys.path.append(os.path.dirname(os.path.dirname(os.path.dirname(os.path.dirname(os.path.dirname(os.path.abspath(__file__)))))))
from webull_helpers import format_date
import pandas as pd


class Event:
    def __init__(self, data):
        self.actual = data.get('actual', None)
        self.comment = data.get('comment', None)
        self.country = data['country']
        self.currency = data['currency']
        self.date = data['date']
        self.time = format_date(self.date)
        self.forecast = data['forecast']
        self.id = data['id']
        self.importance = data['importance']
        self.indicator = data['indicator']
        self.link = data['link']
        self.period = data['period']
        self.previous = data['previous']
        self.scale = data['scale']
        self.source = data['source']
        self.title = data['title']
        self.unit = data['unit']


        self.data_dict = { 
        'actual': self.actual,
        'comment': self.comment,
        'country': self.country,
        'currency': self.currency,
        'date': self.date,
        'time': self.time,
        'forecast': self.forecast,
        'id': self.id,
        'importance': self.importance,
        'indicator': self.indicator,
        'link': self.link,
        'period': self.period,
        'previous': self.previous,
        'scale': self.scale,
        'source': self.source,
        'title': self.title,
        'unit': self.unit
    }
        self.df = pd.DataFrame(self.data_dict)
    def __str__(self):
        return f"Event(id={self.id}, date={self.date}, time={self.time}, title={self.title}, importance={self.importance})"

    def __repr__(self):
        return self.__str__()